from anndata import AnnData

from anndata import read as read_h5ad
from anndata import read_csv, read_excel, read_hdf, read_loom, read_mtx, read_text, read_umi_tools

from .. import __version__

from . import pp
from . import tl
#from . import functions
from . import ct
from . import pl

MOUSE = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', 
        '2', '3', '4', '5', '6', '7', '8', '9','X', 'Y']
HUMAN = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22',
        '2', '3', '4', '5', '6', '7', '8', '9','X', 'Y']

__doc__ = """\
API
===

Import epiScanpy's high-level API as::

   import episcanpy.api as epi

Count Matrices: CT
------------------

Loading data, loading annotations, building count matrices, filtering of lowly covered methylation variables.
Filtering of lowly covered cells.

Load features
~~~~~~~~~~~~~

In order to build a count matrix for either methylation or open chromatin data, loading the segmentation of the genome of interest or the set of features of interest is a prerequirement.

.. autosummary::
   :toctree: .

   ct.load_features
   ct.make_windows
   ct.size_feature_norm
   ct.plot_size_features
   ct.name_features

Reading methylation file
~~~~~~~~~~~~~~~~~~~~~~~~

Functions to read methylation files, extract methylation and buildthe count matrices:
 
 
.. autosummary::
   :toctree: .

   ct.build_count_mtx
   ct.read_cyt_summary
   ct.load_met_noimput

   
Reading open chromatin(ATAC) file
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ATAC-seq specific functions to build count matrices and load data:
 
 
.. autosummary::
   :toctree: .
   
   ct.bld_atac_mtx
   ct.save_sparse_mtx

   
General functions
~~~~~~~~~~~~~~~~~

Functions non -omic specific:

   .. autosummary::
   :toctree: .

   ct.save_sparse_mtx


Preprocessing: PP
------------------

Imputing missing data (methylation), filtering lowly covered cells or variables, correction for batch effect.


.. autosummary::
   :toctree: .

   pp.coverage_cells
   pp.commoness_features
   pp.binarize
   pp.lazy
   pp.load_metadata
   pp.read_ATAC_10x


Methylation matrices
~~~~~~~~~~~~~~~~~~~~

Methylation specific count matrices. 

.. autosummary::
   :toctree: .

   pp.imputation_met
   pp.load_met_noimput
   pp.readandimputematrix


   
Tools: TL
---------

.. autosummary::
   :toctree: .

   tl.rank_features
   tl.silhouette
   tl.lazy
   tl.load_markers
   tl.identify_cluster


Plotting: PL
------------

The plotting module :class:`episcanpy.plotting` largely parallels the ``tl.*`` and a few of the ``pp.*`` functions.
For most tools and for some preprocessing functions, you'll find a plotting function with the same name.

.. autosummary::
   :toctree: .
   pl.prct_overlap
   pl.overlap_heatmap
   pl.silhouette
   pl.silhouette_tot


   
"""



"""Logging and Profiling
"""

import time as time_module
import datetime
#from anndata import logging
from . import settings


_VERBOSITY_LEVELS_FROM_STRINGS = {
    'error': 0,
    'warn': 1,
    'info': 2,
    'hint': 3,
}


def info(*args, **kwargs):
    return msg(*args, v='info', **kwargs)


def error(*args, **kwargs):
    args = ('Error:',) + args
    return msg(*args, v='error', **kwargs)


def warn(*args, **kwargs):
    args = ('WARNING:',) + args
    return msg(*args, v='warn', **kwargs)


def hint(*args, **kwargs):
    return msg(*args, v='hint', **kwargs)


def _settings_verbosity_greater_or_equal_than(v):
    if isinstance(settings.verbosity, str):
        settings_v = _VERBOSITY_LEVELS_FROM_STRINGS[settings.verbosity]
    else:
        settings_v = settings.verbosity
    return settings_v >= v


def msg(*msg, v=4, time=False, memory=False, reset=False, end='\n',
        no_indent=False, t=None, m=None, r=None):
    """Write message to logging output.

    Log output defaults to standard output but can be set to a file
    by setting `sc.settings.log_file = 'mylogfile.txt'`.

    v : {'error', 'warn', 'info', 'hint'} or int, (default: 4)
        0/'error', 1/'warn', 2/'info', 3/'hint', 4, 5, 6...
    time, t : bool, optional (default: False)
        Print timing information; restart the clock.
    memory, m : bool, optional (default: Faulse)
        Print memory information.
    reset, r : bool, optional (default: False)
        Reset timing and memory measurement. Is automatically reset
        when passing one of ``time`` or ``memory``.
    end : str (default: '\n')
        Same meaning as in builtin ``print()`` function.
    no_indent : bool (default: False)
        Do not indent for ``v >= 4``.
    """
    # variable shortcuts
    if t is not None: time = t
    if m is not None: memory = m
    if r is not None: reset = r
    if isinstance(v, str):
        v = _VERBOSITY_LEVELS_FROM_STRINGS[v]
    if v == 3:  # insert "--> " before hints
        msg = ('-->',) + msg
    if v >= 4 and not no_indent:
        msg = ('   ',) + msg
    if _settings_verbosity_greater_or_equal_than(v):
        if not time and not memory and len(msg) > 0:
            _write_log(*msg, end=end)
        if reset:
            try:
                settings._previous_memory_usage, _ = get_memory_usage()
            except:
                pass
            settings._previous_time = time_module.time()
        if time:
            elapsed = get_passed_time()
            msg = msg + ('({})'.format(_sec_to_str(elapsed)),)
            _write_log(*msg, end=end)
        if memory:
            _write_log(format_memory_usage(get_memory_usage()),
                        msg='' if time else msg, end=end)

m = msg  # backwards compat


def _write_log(*msg, end='\n'):
    """Write message to log output, ignoring the verbosity level.

    This is the most basic function.

    Parameters
    ----------
    *msg :
        One or more arguments to be formatted as string. Same behavior as print
        function.
    """
    from .settings import logfile
    if logfile == '':
        print(*msg, end=end)
    else:
        out = ''
        for s in msg:
            out += str(s) + ' '
        with open(logfile, 'a') as f:
            f.write(out + end)


def _sec_to_str(t):
    """Format time in seconds.

    Parameters
    ----------
    t : int
        Time in seconds.
    """
    from functools import reduce
    return "%d:%02d:%02d.%02d" % \
        reduce(lambda ll, b: divmod(ll[0], b) + ll[1:],
               [(t*100,), 100, 60, 60])


#print_memory_usage = logging.print_memory_usage


#get_memory_usage = logging.get_memory_usage


def get_passed_time():
    now = time_module.time()
    elapsed = now - settings._previous_time
    settings._previous_time = now
    return elapsed


def print_version_and_date():
    from . import __version__
    _write_log('Running epiScanpy', __version__, 'on {}.'.format(get_date_string()))


_DEPENDENCIES_NUMERICS = [
    'anndata',  # anndata actually shouldn't, but as long as it's in development
    'numpy',
    'scipy',
    'pandas',
    ('sklearn', 'scikit-learn'),
    'statsmodels',
    ('igraph', 'python-igraph'),
    'louvain']


_DEPENDENCIES_PLOTTING = ['matplotlib', 'seaborn']


def _print_versions_dependencies(dependencies):
    # this is not the same as the requirements!
    for mod in dependencies:
        mod_name = mod[0] if isinstance(mod, tuple) else mod
        mod_install = mod[1] if isinstance(mod, tuple) else mod
        try:
            imp = __import__(mod_name)
            print('{}=={}'.format(mod_install, imp.__version__), end=' ')
        except (ImportError, AttributeError):
            pass
    print()

def print_versions():
    """Versions that might influence the numerical results.

    Matplotlib and Seaborn are excluded from this.
    """
    _print_versions_dependencies(['episcanpy'] + _DEPENDENCIES_NUMERICS)


def print_versions_dependencies_numerics():
    """Dependencies that might influence numerical results (computed data).
    """
    print('Dependencies:', end=' ')
    _print_versions_dependencies(_DEPENDENCIES_NUMERICS)


def print_versions_dependencies_plotting():
    """Dependencies that might influence plots (but not computed data).
    """
    print('Dependencies:', end=' ')
    _print_versions_dependencies(_DEPENDENCIES_PLOTTING)


def print_versions_dependencies_all():
    """All relevant dependencies.
    """
    _print_versions_dependencies(
        _DEPENDENCIES_NUMERICS + _DEPENDENCIES_PLOTTING)


def get_date_string():
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M")



import numpy as np
from scipy.io import mmread
import pandas as pd
import anndata as ad


def read_ATAC_10x(matrix, cell_names='', var_names='', path_file=''):
    """
    Load sparse matrix (including matrices corresponding to 10x data) as AnnData objects.
    read the mtx file, tsv file coresponding to cell_names and the bed file containing the variable names

    Parameters
    ----------
    matrix: sparse count matrix

    cell_names: optional, tsv file containing cell names

    var_names: optional, bed file containing the feature names

    Return
    ------
    AnnData object

    """

    
    mat = mmread(''.join([path_file, matrix]))
    mat = mat.toarray()
    mat = np.matrix(mat.transpose())
    
    with open(path_file+cell_names) as f:
        barcodes = f.readlines() 
        barcodes = [x[:-1] for x in barcodes]
        
    with open(path_file+var_names) as f:
        var_names = f.readlines()
        var_names = ["_".join(x[:-1].split('\t')) for x in var_names]
        
    adata = ad.AnnData(mat, obs=pd.DataFrame(index=barcodes), var=pd.DataFrame(index=var_names))
    adata.uns['omic'] = 'ATAC'
    
    return(adata)



import anndata as ad 
import scanpy.api as sc
from sklearn.metrics import silhouette_score, silhouette_samples

def silhouette(adata_name, cluster_annot, value='X_pca', metric='euclidean',
               key_added=None, copy=False):
    """

    Compute silhouette scores.

    It computes the general silhouette score as well as a silhouette score for every cell according 
    to the cell cluster assigned to it. 

    Parameters
    ----------
    adata_name: AnnData object

    cluster_annot: observational variable corresponding to a cell clustering

    value: measure used to build the silhouette plot (X_pca, X_tsne, X_umap)

    metric: 'euclidean'

    key_added: key to save the computed silhouette scores

    Return
    ------

    general silhouette score in 'uns' of the AnnData object
    individual silhouette scores in 'obs' of the AnnData object



    Credit to sklearn script : 
    https://scikit-learn.org/stable/auto_examples/cluster/plot_kmeans_silhouette_analysis.html#sphx-glr-auto-examples-cluster-plot-kmeans-silhouette-analysis-py
    return score and silhouette plot. Still some work to do to finish the function.
    size=None but you can put 'large' if you want a bigger default figure size
    """
    
    if copy:
      adata_name = adata_name.copy()
      
    X = adata_name.obsm[value]
    cluster_labels = adata_name.obs[cluster_annot]
    n_clusters = len(set(adata_name.obs[cluster_annot]))

    ## also, return sample_silhouette_values as adata.obs['silhouette_samples']
    silhouette_avg = silhouette_score(X, cluster_labels, metric)
    sample_silhouette_values = silhouette_samples(X, cluster_labels, metric)
    
    if key_added:
        adata_name.obs[key_added] = sample_silhouette_values
        adata_name.uns[key_added] = silhouette_avg
    else:
        adata_name.obs['silhouette_samples'] = sample_silhouette_values
        adata_name.uns['silhouette_samples_avg'] = silhouette_avg

    if copy:
        return(adata_name)
    else:
        return()



import matplotlib.pyplot as plt
import anndata as ad 
import scanpy.api as sc
from sklearn.metrics import silhouette_score, silhouette_samples
import matplotlib.cm as cm
import numpy as np

def silhouette(adata_name, cluster_annot, key=None,
              xlabel=None, ylabel=None, title=None, size='large',
               name_cluster=True, name_cluster_pos='left', 
              palette=None, save=None):
    """
    Plot the product of tl.silhouette as a silhouette plot

    Parameters
    ----------

    adata_name: AnnData object

    cluster_annot: observational variable corresponding to a cell clustering

    key: specify name of precomputed silhouette scores if not standard


    Return
    ------

    Silhouette plot

    """

    
    silhouette_avg = silhouette_score(X, cluster_labels, metric)
    sample_silhouette_values = silhouette_samples(X, cluster_labels, metric)
    
    if key!=None:
        sample_silhouette_values = adata_name.obs[key]
        silhouette_avg = adata_name.uns[key]
    else:
        adata_name.obs['silhouette_samples'] = sample_silhouette_values
        sample_silhouette_values = adata_name.obs['silhouette_samples']
        silhouette_avg = adata_name.obs['silhouette_samples']
        
    
    fig, ax1 = plt.subplots(1, 1)
    if size=='large':
        fig.set_size_inches(18, 7)
    #ax1.set_xlim([-0.1, 1])
    ax1.set_ylim([0, len(X) + (n_clusters + 1) * 10])
    y_lower = 10
    
    ## for title argument
    if title != None:
        ax1.set_title(title)
    else:
        ax1.set_title(cluster_annot)
    ## axis label arguments
    if xlabel != None:
        ax1.set_xlabel(xlabel)
    if ylabel == None:
        ax1.set_ylabel(cluster_annot)
    else:
        ax1.set_ylabel(ylabel)

        
    i =0
    for name in list(sorted(set(adata_name.obs[cluster_annot]))):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = \
            sample_silhouette_values[cluster_labels == name]

        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i
        
        if palette != None:
            color = palette[i]
        else:
            color = cm.nipy_spectral(float(i) / n_clusters)
        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                            0, ith_cluster_silhouette_values,
                            facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        #ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))
        if name_cluster:
            if name_cluster_pos == 'left':
                ax1.text((min(sample_silhouette_values)+0.1), y_lower + 0.5 * size_cluster_i, name)
            else:
                ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, name)
            

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples
        i += 1


    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhouette_avg, color="red", linestyle="--")
    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    #ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])
    if save != None:
        plt.savefig('_'.join(['silhouette', save])) 
    plt.show()
    
    print(silhouette_avg)
    return()

def silhouette_tot(adata_name, cluster_annot, value='X_pca', metric='euclidean',
              xlabel=None, ylabel=None, title=None, size='large',
               name_cluster=True, name_cluster_pos='left', 
              palette=None, save=None, key_added=None):
    """
    Both compute silhouette scores and plot it.

    It computes the general silhouette score as well as a silhouette score for every cell according 
    to the cell cluster assigned to it. 

    Parameters
    ----------
    adata_name: AnnData object

    cluster_annot: observational variable corresponding to a cell clustering

    value: measure used to build the silhouette plot (X_pca, X_tsne, X_umap)

    metric: 'euclidean'

    key_added: key to save the computed silhouette scores

    Return
    ------

    general silhouette score in 'uns' of the AnnData object
    individual silhouette scores in 'obs' of the AnnData object
    
    Silhouette plot


    Credit to sklearn script : 
    https://scikit-learn.org/stable/auto_examples/cluster/plot_kmeans_silhouette_analysis.html#sphx-glr-auto-examples-cluster-plot-kmeans-silhouette-analysis-py
    return score and silhouette plot. Still some work to do to finish the function.
    size=None but you can put 'large' if you want a bigger default figure size
    """
    #sc.pl.umap(adata_name, color=cluster_annot)
    #print(list(sorted(set(adata_name.obs[cluster_annot]))))

    X = adata_name.obsm[value]
    cluster_labels = adata_name.obs[cluster_annot]
    n_clusters = len(set(adata_name.obs[cluster_annot]))


    fig, ax1 = plt.subplots(1, 1)
    if size=='large':
        fig.set_size_inches(18, 7)
    #ax1.set_xlim([-0.1, 1])
    ax1.set_ylim([0, len(X) + (n_clusters + 1) * 10])
    y_lower = 10
    
    ## for title argument
    if title != None:
        ax1.set_title(title)
    else:
        ax1.set_title(cluster_annot)
    ## axis label arguments
    if xlabel != None:
        ax1.set_xlabel(xlabel)
    if ylabel == None:
        ax1.set_ylabel(cluster_annot)
    else:
        ax1.set_ylabel(ylabel)

    ## depending of verbosity, print(silhouette_avg)
    ## also, return sample_silhouette_values as adata.obs['silhouette_samples']
    silhouette_avg = silhouette_score(X, cluster_labels, metric)
    sample_silhouette_values = silhouette_samples(X, cluster_labels, metric)
    if key_added:
        adata_name.obs[key_added] = sample_silhouette_values
    else:
        adata_name.obs['silhouette_samples'] = sample_silhouette_values


    i =0
    for name in list(sorted(set(adata_name.obs[cluster_annot]))):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = \
            sample_silhouette_values[cluster_labels == name]

        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i
        
        if palette != None:
            color = palette[i]
        else:
            color = cm.nipy_spectral(float(i) / n_clusters)
        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                            0, ith_cluster_silhouette_values,
                            facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        #ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))
        if name_cluster:
            if name_cluster_pos == 'left':
                ax1.text((min(sample_silhouette_values)+0.1), y_lower + 0.5 * size_cluster_i, name)
            else:
                ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, name)
            

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples
        i += 1


    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhouette_avg, color="red", linestyle="--")
    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    #ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])
    if save != None:
        plt.savefig('_'.join(['silhouette', save])) 
    plt.show()
    
    print(silhouette_avg)
    return()




import numpy as np
import anndata as ad
import pandas as pd

def load_peak_matrix(matrix_file, path=''):
    
    """
    Load existing peak x cell matrix into an anndata object.
    Input existing peak matrix in the format as in Cusanovich et al. 
    First row of the matrix file is considered the header (chr start end annot cell1...cellN)
    Paramters
    ---------
    matrix_file: txt or tsv file with the matrix in shape peaks x cells  
    
    Return
    ------
    AnnData
    """
    
    peak_name = []
    cell_matrix = [] 
    with open(path+matrix_file) as f:
        head = f.readline().split('\t')
        head[len(head)-1] = head[len(head)-1].split("\n")[0]
        for line in f:
            line = line.split('\t')
            line[len(line)-1] = line[len(line)-1].split("\n")[0]
            peak_name.append(line[3]) # for some reason it has rownames
            line2 = []
            for x in line[4:]:
                if x =='0':
                    line2.append(0)
                else:
                    line2.append(1)
            cell_matrix.append(line2)
    cell_names = head[4:]
    
    cell_matrix=np.matrix(cell_matrix)
    cell_matrix = cell_matrix.transpose()
    
    adata = ad.AnnData(cell_matrix,
                   obs=pd.DataFrame(index=cell_names),
                   var=pd.DataFrame(index=peak_name))
    adata.uns['omic'] = 'ATAC'
    
    return(adata)


from ..count_matrix._features import load_features, make_windows, size_feature_norm, plot_size_features, name_features
from ..count_matrix._load_input_file import read_cyt_summary
from ..count_matrix._bld_met_mtx import build_count_mtx
from ..count_matrix._atac_mtx import bld_atac_mtx, save_sparse_mtx
from ..count_matrix._load_met_ct_mtx import load_met_noimput


import scanpy.api as sc
import anndata as ad

def lazy(adata, pp_pca=True, nb_pcs=50, n_neighbors=15, perplexity=30, 
         method='umap', metric='euclidean', min_dist=0.5, spread=1.0, 
         n_components=2, copy=False):
    '''
    Automatically computes PCA coordinates, loadings and variance decomposition, a neighborhood graph of observations,
    t-distributed stochastic neighborhood embedding (tSNE) Uniform Manifold Approximation and Projection (UMAP)
    
    Parameters
    ----------
    adata : :class:`~anndata.AnnData`
        Annotated data matrix.
    pp_pca :  `bool` (default: `True`)
        Computes PCA coordinates before the neighborhood graph
    nb_pcs : `int` (default: 50)
        Number of principal component computed for PCA (and therefore neighbors, tsne and umap)
    n_neighbors : `int` (default: 15)
        The size of local neighborhood (in terms of number of neighboring data points) used for manifold approximation.
        Larger values result in more global views of the manifold, while smaller values result in more local data being
        preserved. In general values should be in the range 2 to 100.
    method : `str` (default: ‘umap’)
        Use ‘umap’ or ‘gauss’, kernel for computing connectivities. Gives very similar results.
    metric : `str` (default: ‘euclidean’)
        A known metric’s name or a callable that returns a distance.
    perplexity : `int` (default: 30)
        The perplexity is related to the number of nearest neighbors that is used in other manifold learning algorithms.
        Larger datasets usually require a larger perplexity. Consider selecting a value between 5 and 50. 
    min_dist : `float` (default: 0.5)
        The effective minimum distance between embedded points. Smaller values will result in a more clustered/clumped
        embedding where nearby points on the manifold are drawn closer together, while larger values will result on a
        more even dispersal of points.
    spread : `float` (default: 1.0)
        The effective scale of embedded points. In combination with `min_dist` this determines how clustered/clumped
        the embedded points are.
    n_components : `int` (default: 2)
        The number of dimensions of the `UMAP` embedding.
    copy : `bool` (default: `False`)
        Return a copy instead of writing to adata.
    
    method='umap',
    metric='euclidean',
    min_dist=0.5,
    spread=1.0, 
    n_components=2
    
    
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    X_pca : `adata.obsm`
        PCA coordinates of data.
    connectivities : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Weighted adjacency matrix of the neighborhood graph of data
        points. Weights should be interpreted as connectivities.
    distances : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Instead of decaying weights, this stores distances for each pair of
        neighbors.
    X_tsne : `np.ndarray` (`adata.obs`, dtype `float`)
        tSNE coordinates of data.
    X_umap : `adata.obsm`
        UMAP coordinates of data.
    
    
    '''
    if copy:
        adata = adata.copy() 
    else:
        adata

    if pp_pca:
        sc.pp.pca(adata, n_comps=nb_pcs)
    
    sc.pp.neighbors(adata,  n_neighbors=n_neighbors, n_pcs=nb_pcs, method=method, metric=metric)
    sc.tl.pca(adata, n_comps=nb_pcs)
    sc.tl.tsne(adata, n_pcs=nb_pcs, perplexity=perplexity)
    sc.tl.umap(adata, min_dist, spread, n_components)
    
    if copy:
        return(adata)
    else:
        None
  



import anndata as ad

def load_metadata(adata, metadata_file, path='', separator=';'):
    """
    Load observational metadata in adata.obs.
    Input metadata file as csv/txt and the adata object to annotate.

    first raw of the metadata file is considered as a header
    first column contain the cell name

    Paramters
    ---------

    adata: initial AnnData object

    metadata_file: csv file containing as a first column the cell names and in the
        rest of the columns any king of metadata to load

    path: pathe to the metadata file

    separator: ';' or "\t", character used to split the columns
    
    Return
    ------
    Annotated AnnData 
    """
    dict_annot = {}
    with open(path+metadata_file) as f:
        head = f.readline().split(separator)
        file = f.readlines()
    for key in head:
        dict_annot[key] = []
    data = [line.split(separator) for line in file]

    for name in adata.obs_names:
        name = name.split('.')[0]
        found = False
        for line in data:
            if name == line[0]:
                i = 0
                for key in head:
                    dict_annot[key].append(line[i])
                    i += 1 
                found = True
                continue
        # if we could not find annotations
        if found == False:
            for key in head:
                dict_annot[key].append('NA')

    for key in head:
        adata.obs[key] = dict_annot[key]

    return()

### function to optimize running time

def find_genes(adata, gtf_file_name, path='', extension=5000,
               key_added='gene_name', feature_coordinates=None, copy=True):
    """
    Given a gtf file, you can match peak coordinates (stored in adata.var_names or
    in a var annotation) to genes.
    The peak annotation has to be written as chr1:20000-20500 or chr1_20000_20500.
    the corresponding gene (if any) will be sotred in a var annotation 
    It extend the search to match a gene to an window of + and - extensions size(5kb
    for example).
    """
    start = time.time()

    # load the gtf file
    gtf_file = []
    with open(gtf_file_name) as f:
        for line in f:
            if line[0] != '#':
                gtf_file.append(line)
    gtf_file = pd.DataFrame([l.split('\t') for l in gtf_file])
    gtf_file.columns = ['Chromosome', 'source', 'gene_type', 'Start', 'End',
                   'NA', 'Strand', 'NA2', 'extra_info'] 
    
    del gtf_file['NA'], gtf_file['NA2']
    
    # extract the variable names
    feature_annot = adata.var
    if feature_coordinates ==None:
        feature_names = adata.var_names.tolist()
    else:
        feature_names = adata.var[feature_coordinates]
    
    # format the feature name
    start_feature = []
    end_feature = []
    chrom_feature = []
    for feature in feature_names:
        if ':' in feature: # if the feature is a Granger coordinate.
            feature2 = feature.split(':')
            w = [int(x) for x in feature2[1].split('-')]
        else:
            feature2 = feature.split('_')
            w = [int(x) for x in feature2[1:]]
        chrom_feature.append(feature2[0][3:])
        start_feature.append(w[0])
        end_feature.append(w[1])
    
    adata.var['Index'] = range(0,len(chrom_feature))
    adata.var['Chromosome'] = chrom_feature
    adata.var['Start'] = start_feature
    adata.var['End'] = end_feature
    adata.var['name_feature'] = adata.var_names.tolist()
    #adata.var['chrom_feature'] = chrom_feature
    adata.var['start_ext'] = [x-extension for x in start_feature]
    adata.var['end_ext'] = [x+extension for x in end_feature]
    #adata.var['Strand'] = len(end_feature)*['+']
    
    
    # match the feature with 
    gtf = pr.PyRanges(gtf_file)
    del gtf_file
    adata_var = pr.PyRanges(chromosomes=adata.var.loc[:,'Chromosome'], #strands=adata.var.loc[:,'strand_feature'],
               starts=adata.var.loc[:,'start_ext'], ends=adata.var.loc[:,'end_ext'])
    
    merge = gtf.join(adata_var, suffix="_ext")
    merge = merge.dfs
    overlap3 = pd.concat([merge[key] for key in merge.keys()])
    overlap3['Index'] = overlap3.index
    overlap4 = overlap3.sort_values(['Chromosome', 'Start_ext', 'End_ext', 'Index'])
     
    print(time.time()-start)
    
    adata.var = adata.var.sort_values(['Chromosome', 'start_ext', 'end_ext'])
    adata_var = pr.PyRanges(adata.var)
    tot_gene_annot = []
    for chrom in list(set(adata.var['Chromosome'])):
        index_gtf = 0
        #next_index = 0
        #curr_adata = adata_var[chrom].df
        overlap3 = pd.concat([merge[key] for key in [(chrom, '+'), (chrom, '-')]])
        overlap3['Index'] = overlap3.index
        overlap3 = overlap3.sort_values(['Chromosome', 'Start_ext', 'End_ext', 'Index'])
        overlap_chrom = overlap3['Start_ext'].tolist()
        #for line_adata in curr_adata[['start_ext']].iterrows():
        for line_adata in adata_var[chrom].df[['start_ext']].iterrows():
            gene_annot = []
            for start_gtf in overlap_chrom[index_gtf:]:
                if start_gtf == line_adata[1][0]:
                    gene_annot.append(overlap3.iloc[index_gtf])
                    index_gtf += 1
                    continue
                else:
                    #index_gtf = next_index
                    break
                
            if gene_annot == []:
                tot_gene_annot.append(('NA'))
            else:
                tot_gene_annot.append(tuple(gene_annot))
        print(chrom, time.time()-start)
    
    
    adata.var[key_added] = tot_gene_annot
    adata.var.sort_values(['Index'])
    print(time.time()-start)
    return(tot_gene_annot, overlap4)


# some technical stuff
import sys
from .utils import check_versions, annotate_doc_types
from ._version import get_versions  # version generated by versioneer

__author__ = ', '.join([
    'Anna Danese',
    'Maria Richter',
])
__email__ = ', '.join([
    'anna.danese@helmholtz-muenchen.de',
    # We don’t need all, the main authors are sufficient.
])
__version__ = get_versions()['version']

check_versions()
annotate_doc_types(sys.modules[__name__], 'episcanpy')
del get_versions, sys, check_versions, annotate_doc_types

# the actual API
from . import tools as tl
from . import preprocessing as pp
from . import count_matrix as ct
from . import plotting as pl
#from . import functions as functions

#from . import datasets

from anndata import AnnData
from anndata import read_h5ad, read_csv, read_excel, read_hdf, read_loom, read_mtx, read_text, read_umi_tools
#from .readwrite import read, read_10x_h5, read_10x_mtx
#from .neighbors import Neighbors
from .settings import set_figure_params

import scanpy as sc


#from ..tools._impute_meth import readandimputematrix
from ..tools._features_selection import rank_features
from ..tools._silhouette import silhouette
from ..tools._recipe import lazy
from ..tools._cell_id import load_markers, identify_cluster
from ..tools._find_genes import top_feature_genes



import scanpy.api as sc
import warnings
from collections import Counter

def load_markers(path, marker_list_file):
    """
    Convert list of known cell type markers from literature to a dictionary
    Input list of known marker genes 
    First row is considered the header
    Paramters
    ---------
    marker_list_file: file with the known marker genes (gene name is the 3rd column)
    
    Return
    ------
    cell_type_markers
    """
    marker_list = []
    cell_type_list = []
    with open(path+marker_list_file) as f:
        head = f.readline()
        for line in f:
            line = line.split("\t")
            line[-1] = line[-1][:-1]
            marker_list.append(line[2].upper())
            cell_type_list.append(line[0])
            
    unique_types = list(set(cell_type_list))
    
    # dictionary: list of marker genes per cell type
    cell_type_markers = {}
    for i in range(0,len(set(cell_type_list))):
        cell_type_markers[unique_types[i]] = []
        for j in range(0, len(cell_type_list)):
            if cell_type_list[j] == unique_types[i]:
                cell_type_markers[unique_types[i]].append(marker_list[j])
                j += 1
                
    return(cell_type_markers)


### load Cusanovich peak/promoter intersections
### need to also set up function to create the file from scratch instead of taking it from Cusanovich 
### (I'm pretty sure I anyway already have a script for that)

def identify_cluster(adata, cell_type, cell_type_markers, peak_promoter_file, gene_name_pos=5, path='', n_peaks_per_cluster=1000):
    """
    Use markers of a given cell type to plot peak openness for peaks in promoters of the given markers
    Input cell type, cell type markers, peak promoter intersections 
    Paramters
    ---------
    cell_type: str, cell type that is to be investigated (must be the same as in the dictionary)
    cell_type_markers: dict, output of load_markers
    peak_promoter_file: tab-separated file that stores information about peak/promoter intersections and gene names for these promoters
    gene_name_pose: int, indicates the column in the peak_promoter_file where the gene name is stored
    path: str, path to the peak_promoter_file
    n_peaks_per_cluster: int, number of peaks per louvain cluster that should be searched for matches with the markers
    
    Return
    ------
    umap depicting peak openness for promoters of marker genes for a given cell type
    """
    if "omic" in adata.uns.keys():
        if adata.uns['omics'] == 'ATAC':
            return(identify_cluster_ATAC(adata, cell_type, cell_type_markers,
                peak_promoter_file, gene_name_pos, path, n_peaks_per_cluster))
        else:
            message = 'the current implementation of this function only allow ATAC-seq data as input'
            warnings.warn(message)
    else:
        message = """the current implementation of this function only allow ATAC-seq data as input.\n
        Additionally, annData uns['omics'] is not specified."""
        warnings.warn(message) 

    return()



def identify_cluster_ATAC(adata, cell_type, cell_type_markers, peak_promoter_file, gene_name_pos=5, path='', n_peaks_per_cluster=1000):
    
    peak_name = []
    gene_name = []
    with open(path+peak_promoter_file) as f:
        head = f.readline()
        for line in f:
            line = line.split("\t")
            peak_name.append(line[0])
            gene_name.append(line[gene_name_pos].upper())
            
    matching_peaks = []
    matching_gene_name = []
    for i in range(0,len(peak_name)):
        # check if peak is in the adata object and if the according gene name is a marker
        if peak_name[i] in list(adata.var_names) and gene_name[i] in cell_type_markers[cell_type]:
            matching_peaks.append(peak_name[i])
            matching_gene_name.append(gene_name[i])
    # get the cell type names per matching peak
    cell_type_peaks = {}
    cell_type_peaks[cell_type] = []
    for j in range(0,len(matching_peaks)):
        if matching_gene_name[j] in cell_type_markers[cell_type]:
            cell_type_peaks[cell_type].append(matching_peaks[j]+"_"+matching_gene_name[j])
            
    # make the peaks unique
    cell_type_peaks[cell_type] = list(set(cell_type_peaks[cell_type]))
        
        
    cell_type_peak = []
    cell_type_gene = []
    for elem in cell_type_peaks[cell_type]:
        ctype = elem.split("_")
        cell_type_peak.append(ctype[0]+"_"+ctype[1]+"_"+ctype[2])
        cell_type_gene.append(ctype[3])
    
    sc.tl.rank_genes_groups(adata, groupby="louvain", n_genes=n_peaks_per_cluster)

    # top 100 to 1000 ranked peaks per louvain group
    ATAC_ranking_dict = {}
    for group in range(0, len(set(list(adata.obs["louvain"])))):
        ATAC_ranking_dict[str(group)] = []
        for idx in range(0,n_peaks_per_cluster):
            ATAC_ranking_dict[str(group)].append(list(adata.uns["rank_genes_groups"]['names'])[idx][group])
        
    sig_markers = []
    sig_peak = []
    for group in range(0, len(set(list(adata.obs["louvain"])))):
        for elem in ATAC_ranking_dict[str(group)]:
            if elem in cell_type_peak:
                sig_markers.append(elem+":"+str(group))
                sig_peak.append(elem)                
            
    if len(sig_markers) > 200:
        print("WARNING: found more than 200 marker peaks for your chosen cell type. Consider choosing a smaller number for n_peaks_per_cluster to reduce running time and get more significant results.")
    
    group_contribution = []
    for element in sig_markers:
        group = element.split(":")[1]
        group_contribution.append(group)

    most_prominent = max(set(group_contribution), key = group_contribution.count) 
    
    print("Differentially open peaks in promoters for known",cell_type,"marker genes are most prominently found in louvain group",most_prominent)
    
    # only take unique peaks (peaks can be among top markers in more than one group)
    unique_sig_peaks = []
    sig_gene_names = []
    for idx,name in enumerate(cell_type_peak):
        if name in sig_peak:
            sig_gene_names.append(cell_type_gene[idx])
            unique_sig_peaks.append(cell_type_peak[idx])
        
    return(sc.pl.umap(adata, color=unique_sig_peaks, title=sig_gene_names))   


from ..plotting._heatmap import overlap_heatmap, prct_overlap
#from ..plotting._features import plot_rank_features
from ..plotting._silhouette import silhouette, silhouette_tot
from ..plotting._histograms import cluster_composition



# right now it is the same script as in sacnpy. However I need to change it 
# because the figure settings need tobe adapted for 
# the different omics

"""Settings
"""

verbosity = 1
"""Set global verbosity level.
Level 0: only show 'error' messages.
Level 1: also show 'warning' messages.
Level 2: also show 'info' messages.
Level 3: also show 'hint' messages.
Level 4: also show very detailed progress.
Level 5: also show even more detailed progress.
etc.
"""

plot_suffix = ''
"""Global suffix that is appended to figure filenames.
"""

file_format_data = 'h5ad'
"""File format for saving AnnData objects.
Allowed are 'txt', 'csv' (comma separated value file) for exporting and 'h5ad'
(hdf5) for lossless saving.
"""

file_format_figs = 'pdf'
"""File format for saving figures.
For example 'png', 'pdf' or 'svg'. Many other formats work as well (see
`matplotlib.pyplot.savefig`).
"""

autosave = False
"""Save plots/figures as files in directory 'figs'.
Do not show plots/figures interactively.
"""

autoshow = True
"""Show all plots/figures automatically if autosave == False.
There is no need to call the matplotlib pl.show() in this case.
"""

writedir = './write/'
"""Directory where the function scanpy.write writes to by default.
"""

cachedir = './cache/'
"""Default cache directory.
"""

figdir = './figures/'
"""Directory where plots are saved.
"""

max_memory = 15
"""Maximal memory usage in Gigabyte.
Is currently not well respected....
"""

n_jobs = 1
"""Default number of jobs/ CPUs to use for parallel computing.
"""

logfile = ''
"""Name of logfile. By default is set to '' and writes to standard output."""

categories_to_ignore = ['N/A', 'dontknow', 'no_gate', '?']
"""Categories that are omitted in plotting etc.
"""

_frameon = True
"""See set_figure_params.
"""


# --------------------------------------------------------------------------------
# Functions
# --------------------------------------------------------------------------------


def set_figure_params(scanpy=True, dpi=80, dpi_save=150, frameon=True, vector_friendly=True,
                      color_map=None, format='pdf', transparent=False, ipython_format='png2x'):
    """Set resolution/size, styling and format of figures.
    Parameters
    ----------
    scanpy : `bool`, optional (default: `True`)
        Init default values for ``matplotlib.rcParams`` suited for Scanpy.
    dpi : `int`, optional (default: `80`)
        Resolution of rendered figures - this influences the size of figures in notebooks.
    dpi_save : `int`, optional (default: `150`)
        Resolution of saved figures. This should typically be higher to achieve
        publication quality.
    frameon : `bool`, optional (default: `True`)
        Add frames and axes labels to scatter plots.
    vector_friendly : `bool`, optional (default: `True`)
        Plot scatter plots using `png` backend even when exporting as `pdf` or `svg`.
    color_map : `str`, optional (default: `None`)
        Convenience method for setting the default color map.
    format : {'png', 'pdf', 'svg', etc.}, optional (default: 'pdf')
        This sets the default format for saving figures: `file_format_figs`.
    transparent : `bool`, optional (default: `True`)
        Save figures with transparent back ground. Sets
        `rcParams['savefig.transparent']`.
    ipython_format : list of `str`, optional (default: 'png2x')
        Only concerns the notebook/IPython environment; see
        `IPython.core.display.set_matplotlib_formats` for more details.
    """
    try:
        import IPython
        IPython.core.display.set_matplotlib_formats(ipython_format)
    except:
        pass
    from matplotlib import rcParams
    global _vector_friendly
    _vector_friendly = vector_friendly
    global file_format_figs
    file_format_figs = format
    if dpi is not None:
        rcParams['figure.dpi'] = dpi
    if dpi_save is not None:
        rcParams['savefig.dpi'] = dpi_save
    if transparent is not None:
        rcParams['savefig.transparent'] = transparent
    if scanpy:
        from .plotting._rcmod import set_rcParams_scanpy
        set_rcParams_scanpy(color_map=color_map)
    global _frameon
    _frameon = frameon



# ------------------------------------------------------------------------------
# Private global variables & functions
# ------------------------------------------------------------------------------

_vector_friendly = False
"""Set to true if you want to include pngs in svgs and pdfs.
"""

_low_resolution_warning = True
"""Print warning when saving a figure with low resolution."""

def _set_start_time():
    from time import time
    return time()

_start = _set_start_time()
"""Time when the settings module is first imported."""

_previous_time = _start
"""Variable for timing program parts."""

_previous_memory_usage = -1
"""Stores the previous memory usage."""


def _is_run_from_ipython():
    """Determines whether run from Ipython.
    Only affects progress bars.
    """
    try:
        __IPYTHON__
        return True
    except NameError:
        return False



import anndata as ad
import matplotlib.pyplot as plt
import pandas as pd

def cluster_composition(adata, cluster, condition, xlabel='cell cluster',
                        ylabel='cell count', title=None, save=False):
    """
    """
    

    contingency_table = pd.crosstab(
        adata.obs[condition],
        adata.obs[cluster],
        margins = True
    )

    counts = []
    p_part = []
    index = 0
    categories = sorted(list(set(adata.obs[cluster])))
    for n in sorted(set(adata.obs[condition])):
        #counts.append()
        p_part.append(plt.bar(categories, contingency_table.iloc[index][0:-1].values))
        index += 1

    #Plots the bar chart
    #plt.figsize(figsize=[6.4, 4.8])
    plt.legend(tuple([p[0] for p in p_part]), tuple(sorted(set(adata.obs[condition]))))
    plt.xlabel(xlabel, )
    plt.ylabel(ylabel)
    plt.title(title)
    
    
    if save!=False:
        
        if (save==True) or (save.split('.')[-1] not in ['png', 'pdf']):
            plt.savefig('cluster_composition.png', dpi=300, bbox_inches="tight")
        else:
            plt.savefig('_'.join(['cluster_composition',save]), #format=save.split('.')[-1],
                        dpi=300, bbox_inches="tight")
            
    plt.show()


from ..count_matrix._features import load_features, make_windows, size_feature_norm, plot_size_features, name_features
from ..count_matrix._load_input_file import read_cyt_summary
from ..count_matrix._bld_met_mtx import build_count_mtx
from ..count_matrix._atac_mtx import bld_atac_mtx, save_sparse_mtx
from ..count_matrix._load_met_ct_mtx import load_met_noimput
#from ..count_matrix._bld_met_mtx import methylation_level, write_not_sparse_meth, extract_methylation, extract_feature_names
#from ..count_matrix._bld_met_mtx import read_meth_fileCG, read_meth_fileCH, read_meth_file, filter_and_average_features
#from ..count_matrix._bld_met_mtx import filter_and_average_features_chrm, prep_methlevels, write_list, write_methlevel

#from ..count_matrix._read_meth_file import read_methylation_file
#from ..count_matrix._extract import extract_CG, extract_CH



import pandas as pd
import seaborn as sns
import numpy as np


def prct_overlap(adata, key_1, key_2, norm=False, ax_norm="row", sort_index=False):
    """
    % or cell count corresponding to the overlap of different cell types 
    between 2 set of annotations/clusters. 

    Parameters
    ----------

    adata: AnnData objet

    key_1: observational key corresponding to one cell division/ one set of clusters

    key_2: bservational key corresponding to one cell division/ one set of clusters

    norm: normalise the ratio to the cell numbers given the total number of cells per
    cluster in key_1

    Return
    ------

    Table containing the ratio of cells within a cluster
    
    """
    
    data_1 = adata.obs[key_1].tolist()
    data_2 = adata.obs[key_2].tolist()
    
    count = {k:[] for k in list(set(data_1))}
    #count = {k:[] for k in sorted(list(set(data_1)))}
    i = 0
    for index in data_1:
        count[index].append(data_2[i])
        i += 1
    
    total_matrix = []
    for key, value in count.items():
        value = sorted(value)
        curr_key_list = []
        for element in sorted(list(set(data_2))):
            curr_count = 0
            for v in value:
                if element == v: 
                    curr_count += 1
            curr_key_list.append(curr_count)
        curr_sum = sum(curr_key_list) 
        #total_matrix.append([x/curr_sum for x in curr_key_list])
        total_matrix.append(curr_key_list)
        
    if norm and ax_norm == "row":
        total_matrix = []
        for key, value in count.items():
            value = sorted(value)
            curr_key_list = []
            for element in sorted(list(set(data_2))):
                curr_count = 0
                for v in value:
                    if element == v: 
                        curr_count += 1
                curr_key_list.append(curr_count)
            curr_sum = sum(curr_key_list) 
            total_matrix.append([x/curr_sum for x in curr_key_list])
        
    elif norm:
        print("""error in the argument ax_norm or it is col and 
              I haven't figure out how to make it for mow.
              , here is the heatmap with no normalisation""")
            
    if sort_index:
        data_heatmap = pd.DataFrame(data=np.matrix(total_matrix),
                                    index=list(set(data_1)),
                                    columns=sorted(list(set(data_2)))).sort_index()
    else:
        data_heatmap = pd.DataFrame(data=np.matrix(total_matrix),
                                index=list(set(data_1)),
                                columns=sorted(list(set(data_2))))
    
    return(data_heatmap)


def overlap_heatmap(adata, key_1, key_2, color="Blues", norm=False, ax_norm="row", sort_index=False):
    """

    Heatmap of the cluster correspondance between 2 set of annaotations. 

    Parameters
    ----------

    adata: AnnData objet

    key_1: observational key corresponding to one cell division/ one set of clusters

    key_2: bservational key corresponding to one cell division/ one set of clusters
   
    norm: normalisation on the total number of cells per row/column or not.

    ax_norm: normalistion (if norm=True) on row or col
    
    colors: gradient displayed in the heatmap. "YlGnBu", "BuPu", "Greens", "Blues"

    Return
    ------

    Heatmap.


    """
    #  I am not 100 percent convinced with the normalisation right now. I prefer the visual
    # as simple cell count. Despite the bias torward large cell clusters.
   
    data_heatmap = prct_overlap(adata, key_1, key_2, norm, ax_norm, sort_index)
    
    return(sns.heatmap(data_heatmap, xticklabels=True, yticklabels=True, cmap=color))



import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.axes as pltax
import numpy as np
import anndata as ad

def commonness_features(adata, threshold=None, bw=0.5, key_added=None, xlabel=None, title=None, save=None):
    """
    Display how often a feature is measured as open (for ATAC-seq).
    Distribution of the feature commoness in cells.
    """
    if key_added == None:
        key_added='commonness'
    
    common = np.sum(adata.X, axis=0).tolist()
    if len(common) == 1:
        common = [item for sublist in common for item in sublist]
    adata.var[key_added] = common

    if threshold != None:
        plt.axvline(x=threshold, color='r')
    bw_param = bw
    sns.set_style('whitegrid')
    #sns.kdeplot(np.array(adata.var[key_added]), bw=bw_param)
    
    plot = sns.distplot(common, hist=False, kde=True,
                    bins=int(80), color = 'darkblue', 
                    hist_kws={'edgecolor':'black'},
                    kde_kws={'linewidth': 1})
    if xlabel ==None:
        plt.xlabel('cells sharing a feature')
    else:
        plt.xlabel(xlabel)
        
    
    if title !=None:
        plt.title(title)
    
    fig = plot.get_figure()
    if save!= None:
        fig.savefig(save)
    plt.show()

    adata.var[key_added] = common
    
def coverage_cells(adata, bins=50, key_added=None, xlabel=None, ylabel=None, title=None, save=None):
    """
    Histogram of the number of features with an open peak (in the case of ATAC-seq data)
    for every cells.
    """
    if key_added == None:
        key_added='sum_peaks'
    # make sum peaks
    
    sum_peaks = np.sum(adata.X, axis=1)
    if len(sum_peaks) == 1:
        sum_peaks = sum_peaks.tolist()
        sum_peaks = [item for sublist in sum_peaks for item in sublist]
    adata.obs[key_added] = sum_peaks
    
    # number of peaks in a cell
    np.histogram(adata.obs[key_added])
    
    
    # plotting parameters
    if xlabel ==None:
        plt.xlabel('number of features')
    else:
        plt.xlabel(xlabel)
        
    if ylabel ==None:
        plt.ylabel('number of cells')
    else:
        plt.ylabel(ylabel)
    
    if title !=None:
        plt.title(title)
        
    fig = plt.hist(adata.obs[key_added], bins)
    
    #fig = plot.get_figure()
    if save!= None:
        plt.savefig(save)
    plt.show()
    
    
def binarize(adata, copy=False):
    """convert into a binary matrix"""
    threshold, upper, lower = 1.0, 1.0, 0.0
    admatrix = adata.X
    admatrix = np.where(admatrix>threshold, upper, lower)
    if copy:
        adata2 = adata.copy()
        adata2.X = admatrix
        return(adata2)
    else:
        adata.X = admatrix
        



# chromosomes for 2 principal species. If you work with another genome
# the chromosomes will have to be specified
# mitochondrial genome not included
HUMAN = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '2',
         '20', '21', '22', '3', '4', '5', '6', '7', '8', '9', 'X', 'Y']
MOUSE = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '2',
         '3', '4', '5', '6', '7', '8', '9', 'X', 'Y']

def load_features(file_features, chromosomes=HUMAN, path="", sort=False):
    """
    The function load features is here to transform a bed file into a usable 
    set of units to measure methylation levels. 
    It has to be a bed-like file. You also need to specify the chromosomes you 
    use as a list of characteres like ['1', '7', '8', '9', 'X', 'Y', 'M'].
    The chromosome list you give as input can be not ordered.
    If you don't specify the chromosomes, the default is the human genome
    (including X, Y and mitochondrial DNA).
    THE BED FILE need to be sorted !! (Maybe I should add an option to sort the
    file).
    
    The output is a dictionary where the keys are chromosomes and the value is
    a list containing [start, end, name] for every feature extracted.
    
    Parameters
    ----------
    file_features:
        the names of the bed file you want to load.
    chromosomes:
        chromosomes corresponding to the bed file. 
        If not specified, it's human by default
    path:
        if you want to specify the path where your bed file is. 
    sort: 
        if True, the bed file is sorted based on starting coordinates.
    """
    features_chrom = {}
    for c in chromosomes:
        features_chrom[c] = []
    with open(path + file_features) as features:
        for line in features:
            line = line.split("\t")
            if line[0][3:] in chromosomes:
                features_chrom[line[0][3:]].append([int(line[1]), int(line[2]), line[3]])
                
#    for c in chromosomes:
#        if features_chrom[c] ==  []:
#            del features_chrom[c]
#            
    if sort == True:
        for c in chromosomes:
            sorted(features_chrom[c], key=lambda x: x[0])
            
    return(features_chrom)

def make_windows(size, chromosomes=HUMAN, max_length=1000000000):
    """
    Generate windows/bins of the given size for the appropriate genome (default
    choice is human). 
    
    Parameters
    ----------
    size:
        size of the window you want
    chromosomes:
        Chromosomes of the species you are analysing
    max_length:
        maximum length given for every chromosome
    """
    features_chrom = {}
    start = range(0, max_length - size, size)
    end = range(size, max_length, size)
    for c in chromosomes:
        features_chrom[c] = [[start[i], end[i], ''.join(["chr_", c, "_", str(i)])] for i in range(len(end))]
    return(features_chrom)


def size_feature_norm(loaded_feature, size):
    """If the features loaded are too smalls or of different sizes, 
    it is possible to normalise them to a unique given size by extending the 
    feature coordinate in both directions.

    Parameters
    ----------
    loaded_feature: loaded feature to normalise

    size: desired size of the feature

    Return
    ------
    Update the input features

    """

    for key in loaded_feature.keys():
        for i in range(len(loaded_feature[key])):
            length = loaded_feature[key][i][1] - loaded_feature[key][i][0]
            add_length = size - length
            if add_length%2 == 1:
                loaded_feature[key][i][0] = loaded_feature[key][i][0] - add_length//2 -1
                loaded_feature[key][i][1] = loaded_feature[key][i][1] + add_length//2
            else:
                loaded_feature[key][i][0] = loaded_feature[key][i][0] - add_length//2
                loaded_feature[key][i][1] = loaded_feature[key][i][1] + add_length//2
                
def plot_size_features(loaded_feature, bins=50, return_length=False):
    """
    Plot the different feature sizes in an histogram. 
    """
    length = []
    for key in loaded_feature.keys():
        for line in loaded_feature[key]:
            length.append(line[1] - line[0])
        
    np.histogram(length)
    plt.hist(length, bins)
    plt.show
    if return_length:
        return(length)
    
    
def name_features(loaded_features):
    """
    Extract the names of the loaded features, specifying the chromosome they originated from.
    It also contain the feature coordinates and an unique identifier.
    """
    feat_names = []
    i = 0
    for c in loaded_features.keys():
        for name in loaded_features[c]:
            add_name = '_'.join(['chr', c, name[-1].rstrip('\n'), str(i)])
            if add_name[-1] =='\n':
                add_name = add_name[:-1]
            feat_names.append(add_name)
            i += 1
    return(feat_names)



import numpy as np
import anndata as ad

def load_met_noimput(matrix_file, path='', save=False):
    """
    read the raw count matrix and convert it into an AnnData object.
    write down the matrix as .h5ad (AnnData object) if save = True.
    Return AnnData object
    """
    matrix = []
    cell_names = []
    feature_names = []
    with open('../../enhancer_c1_CG_paper_unique_corrected2.txt') as f:
        line = f.readline()[:-2].split('\t')
        if line[0] == 'sample_name':
            feature_names = line[1:-1]
        else:
            matrix.append(line[1:])
            cell_names.append(line[0])
        if matrix == []:
            line = f.readline()[:-2].split('\t')
            matrix.append(line[1:])
            cell_names.append(line[0])
        for line in f:
            line = line[:-2].split('\t')
            matrix.append(line[1:])
            cell_names.append(line[0])

    matrix = np.array(matrix)
    
    if feature_names != []:
        adata = ad.AnnData(matrix, obs=pd.DataFrame(index=cell_names),  var=pd.DataFrame(index=feature_names))
    else:
        adata = ad.AnnData(matrix, obs=pd.DataFrame(index=cell_names))
    
    adata.uns['omic'] = 'methylation'
    
    if save:
        adata.write("".join([".".split(matrix_file)[0],'.h5ad']))
        
    return(adata)



from ..preprocessing._extract import extract_CG, extract_CH
from ..preprocessing._readimpute import load_met_noimput, imputation_met
from ..preprocessing._readimpute import readandimputematrix
from ..preprocessing._quality_control import coverage_cells, commonness_features, binarize
from ..preprocessing._recipe import lazy
from ..preprocessing._metadata import load_metadata
from ..preprocessing._load_matrix import read_ATAC_10x
from ..preprocessing._load_atac import load_peak_matrix



#import pysam
import bamnostic as bs
import numpy as np
import anndata as ad
import pandas as pd

MOUSE = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', 
        '2', '3', '4', '5', '6', '7', '8', '9','X', 'Y']
HUMAN = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22',
        '2', '3', '4', '5', '6', '7', '8', '9','X', 'Y']

def bld_atac_mtx(list_bam_files, loaded_feat, output_file_name=None,
    path=None, writing_option='a', header=None, mode='rb',
    check_sq=True, chromosomes=HUMAN):
    """
    Build a count matrix one set of features at a time. It is specific of ATAC-seq data.
    It curently do not write down a sparse matrix. It writes down a regular count matrix
    as a text file. 
    
    Parameters
    ----------

    list_bam_files: input must be a list of bam file names. One for each cell to 
        build the count matrix for

    loaded_feat: the features for which you want to build the count matrix
        
    output_file_name: name of the output file. The count matrix that will be written
        down in the current directory. If this parameter is not specified, 
        the output count amtrix will be named 'std_output_ct_mtx.txt'

    path: path where to find the input file. The output file will be written down
    in your current directory, it is not affected by this parameter.

    writing_option: standard writing options for the output file. 'a' or 'w'
        'a' to append to an already existing matrix. 'w' to overwrite any 
        previously exisiting matrix. 
        default: 'a'

    header: if you want to write down the feature name specify this argument.
        Input must be a list.

    mode: bamnostic argument 'r' or 'w' for read and write 'b' and 's' for bam or sam
        if only 'r' is specified, bamnostic will try to determine if the input is 
        either a bam or sam file.

    check_sq: bamnostic argument. when reading, check if SQ entries are present in header

    chromosomes: chromosomes of the species you are considering. default value
        is the mouse genome (not including mitochondrial genome).
        MOUSE = '1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19',
                '2', '3', '4', '5', '6', '7', '8', '9','X', 'Y']

    Return
    ------
    It does not return any object. The function write down the desired count
    matrix in a txt file

    """
        
    if output_file_name==None:
        output_file_name='std_output_ct_mtx.txt'


    if path==None:
        path=''
    
    # open file to write
    output_file = open(path+output_file_name, writing_option)
    # write header if specified
    if header != None:
        output_file.write('sample_name\t')
        for feature in header:
            output_file.write(feature)
            output_file.write('\t')
        output_file.write('\n')
    # close file to write   
    output_file.close()

    # start going through the bam files
    for name_file in list_bam_files[0:]:

        ## important variables for output
        index_feat = {key: 0 for key in chromosomes}    
        val_feat = {key: [0 for x in range(len(loaded_feat[key]))] for key in chromosomes}
    
        ## PART 1 read the bam file
        keep_lines = []
        samfile = bs.AlignmentFile(path+output_file_name, mode="rb", check_sq=False)
        for read in samfile.fetch(until_eof=True):
            line = str(read).split('\t')
            if line[2] in chromosomes:
                keep_lines.append(line[2:4])
            ### print -- output
        print(name_file, len(keep_lines), 'mapped reads')
        samfile.close()
        
        ## PART2 reads that fall into 
        for element in keep_lines:
            ## 2 things per line:
            chrom = element[0]
            read_pos = int(element[1])
            max_value_index = len(loaded_feat[chrom])
            ## I want to check if the read map to a feature in the same chrom
            pointer_feat_pos = index_feat[chrom]
            for feat_pos in loaded_feat[chrom][pointer_feat_pos:]:
                pointer_feat_pos += 1
                # Go through all features that are smaller than the read position
                if read_pos > feat_pos[1]:
                    continue
                # if read_pos fall in a feature
                elif read_pos > feat_pos[0]:
                    # Update the pointer for the next read if the pointer isn't out of range
                    if pointer_feat_pos < max_value_index:
                        index_feat[chrom] = pointer_feat_pos
                        val_feat[chrom][pointer_feat_pos] += 1
                    else:
                        index_feat[chrom] = max_value_index
                    # Check the following features without updating the pointer. 
                    break
                else:
                    break
     
            for feat_pos in loaded_feat[chrom][pointer_feat_pos:]:
                # +1 if reads fall into more than one feature
                if feat_pos[0] < read_pos:
                    val_feat[chrom][pointer_feat_pos] += 1
                    pointer_feat_pos += 1
                # if read_pos > start position of the new feature break
                elif read_pos < feat_pos[0]:
                    break
                else:
                    print('error')
                    break
                    
        # PART 3
        # open
        output_file = open(path+output_file_name, 'a')
        # write down the result of the cell
        output_file.write(name_file)
        output_file.write('\t')
        for chrom in chromosomes:
            output_file.write('\t'.join([str(p) for p in val_feat[chrom]]))
            output_file.write('\t')
        output_file.write('\n')
        #close
        output_file.close()
    
def read_mtx_bed(file_name, path='', omic='ATAC'):
    """
    read this specific matrix format. It is the standard output of bedtools when you merge bam files.
    """
    peak_name = []
    cell_matrix = [] 
    with open(path+file_name) as f:
        head = f.readline().split('\t')
        head[len(head)-1] = head[len(head)-1].split("\n")[0]
        for line in f:
            line = line.split('\t')
            line[len(line)-1] = line[len(line)-1].split("\n")[0]
            peak_name.append(line[3]) # for some reason it has rownames
            cell_matrix.append([int(x) for x in line[4:]])
    cell_names = head[4:]
    cell_matrix=np.matrix(cell_matrix)
    cell_matrix = cell_matrix.transpose()
    adata = ad.AnnData(cell_matrix,
                   obs=pd.DataFrame(index=cell_names),
                   var=pd.DataFrame(index=peak_name))
    if omic != None:
        adata.uns['omic'] = omic
    return(adata)

def save_sparse_mtx(initial_matrix, output_file='.h5ad', path='', omic='ATAC', bed=False, save=True):
    """
    Convert regular atac matrix into a sparse matrix:

    Parameters
    ----------

    initial_matrix: the name of the regular count matrix for any given omic to load and convert into 
        a sparse matrix (actually, AnnData object)

    output_file: name of the output file for the AnnData object. Standard output if nothing is specified
        is the name of the inoput file with .h5ad extension

    path: path to the input count matrix. The AnnData object is written in the current directory, 
        not the location specified in path.

    omic: 'ATAC', 'RNA' or 'methylation' are the 3 currently recognised omics in epiScanpy. 
        However, other omic name can be accepted but are not yet recognised in other functions.
        default: 'ATAC'

    bed: boolean. If True it consider another input format (bedtools output format for count matrices)

    save: boolean. If True, the sparse matrix is saved as h5ad file. Otherwise it is simply return.

    Return
    ------

    It returns the loaded matrix as an AnnData object.
    """
    head = None
    data = []
    cell_names = []

    # choice between 2 different input count matrix formats
    if bed == True:
        adata = read_mtx_bed(file_name, path, omic)
    else:
        # reading the non sparse file
        with open(path+initial_matrix) as f:
            first_line = f.readline()
            first_line = first_line[:-3].split('\t')
            if first_line[0] == 'sample_name':
                head = first_line
            else:
                cell_names.append(first_line[0])
                data = [[int(l) for l in first_line[1:]]]
            file = f.readlines()
    
        for line in file:
            line = line[:-3].split('\t')
            cell_names.append(line[0])
            data.append([int(l) for l in line[1:]])

        # convert into an AnnData object
        if head != None:
            adata = ad.AnnData(np.array(data), obs=pd.DataFrame(index=cell_names), var=pd.DataFrame(index=head[1:]))
        else:
            adata = ad.AnnData(np.array(data), obs=pd.DataFrame(index=cell_names))
        
        if omic != None:
            adata.uns['omic'] = omic

    # writing the file as h5ad --> sparse matrix with minimum annotations
    if save ==True:
        if output_file=='.h5ad':
            output_file = "".join([initial_matrix.split('.')[0], output_file])
        
        adata.write(output_file)

    return(adata)
    



from scanpy.api.tl import rank_genes_groups
import warnings

def rank_features(adata, groupby, omic=None, use_raw=True, groups='all', reference='rest', n_genes=100,
                     rankby_abs=False, key_added=None, copy=False, method='',
                     corr_method='benjamini-hochberg', **kwds):

    """
    It is a wrap-up function of scanpy sc.tl.rank_genes_groups function. For more information see 
    Scanpy documentations.

    Here, we optimised the default features for the different kind of omics epiScanpy is analysing.
    Parameters for rankby_abs, method and corr_method are fixed byt the omic (if not specified otherwise)

    In case you want to change all the settings, it is advise to directly use the Scanpy function 
    or to specify the omic parameter as False. 

    If the omic of the input AnnData object is not specified (or incorrect), you can add it in omic 
    (either 'RNA', 'ATAC' or 'methylation'). If the omic of the current matrix is not known by
    epiScanpy but you wnat to use settings of a known omic, specify which omic as a parameter. 

    """
    if omic == None:
    	if 'omic' in adata.uns.keys():
    		omic = adata.uns['omic']
    	else:
    		warning.warn("""Attention: no omic specified. We used default settings of the original Scanpy function\n
    			When the parameters where not specified in input""")
    		omic = 'RNA'

    if (method!='') and (adata.uns['omic'] != 'methylation'):
        method='t-test_overestim_var'
    else:
        method='t-test'
    if omic == 'methylation':
    	if copy==False:
    		sc.tl.rank_genes_groups(adata=adata, groupby=groupby, use_raw=use_raw,
                groups=groups, reference=reference, n_genes=n_genes,
    			rankby_abs=True, key_added=key_added, copy=False, method='t-test', corr_method='benjamini-hochberg')
    		return()
    	else:
    		adata2 = sc.tl.rank_genes_groups(adata=adata, groupby=groupby, use_raw=use_raw,
                groups=groups, reference=reference, n_genes=n_genes,
    			rankby_abs=True, key_added=key_added, copy=True, method='t-test', corr_method='benjamini-hochberg')
    		return(adata2)
    else:
    	if copy==False:
    		sc.tl.rank_genes_groups(adata=adata, groupby=groupby, use_raw=use_raw,
                            groups=groups, reference=reference, n_genes=n_genes,
                     		rankby_abs=rankby_abs, key_added=key_added, copy=False, method=method,
                     		corr_method=corr_method, **kwds)
    		return()
    	else:
    		adata2 = sc.tl.rank_genes_groups(adata=adata, groupby=groupby, use_raw=use_raw,
                            groups=groups, reference=reference, n_genes=n_genes,
                     		rankby_abs=rankby_ab, key_added=key_added, copy=True, method=method,
                            corr_method=corr_method, **kwds)
    		return(adata2)





def read_methylation_file(sample_name, meth_type, head, path, chromosome):
    """
    Read file from which you want to extract the methylation level and
    (assuming it is like the Ecker/Methylpy format) extract the number of
    methylated read and the total number of read for the cytosines covered and
    in the right genomic context (CG or CH)
    Parameters
    ----------
    sample_name:
        name of the file to read to extract key information.
    meth_type:
        CG, CH or not specified
    head: if there is header that you don't want to read. An annotation in the
        file you are reading. The default value is the Methylpy/Ecker header
    path: path of the to access the file of the sample you want to read. 
    chromosome: chromosomes if the species you are considering. default value
        is the human genome (including mitochondrial and sexual chromosomes)
    """
    # open the sample file
    with open(path+sample_name) as sample:
        
        # initialize the variables
        reduced_cyt = {} # to store cyt for each chrom (intermed output)
        for c in chromosome:
            reduced_cyt[c] = []

        # specify the cytosine context you want to consider:
        if meth_type == 'CG':
            extract_CG(sample, reduced_cyt, head)
        elif meth_type == 'CH':
            extract_CH(sample, reduced_cyt, head)
        elif meth_type == '':
            for line in sample:
                line = line.split("\t")
                if line != head:
                    reduced_cyt[line[0]].append([int(line[1]),int(line[-3]), int(line[-2])])
        else:
            print ("error in the choice of methylation context")
            return()
    return(reduced_cyt)



import sys, os
if sys.version_info < (3,):
    sys.exit('epiScanpy requires Python >= 3.5')
from pathlib import Path
from setuptools import setup, find_packages
import versioneer

try:
    from episcanpy import __author__, __email__
except ImportError:  # Deps not yet installed
    __author__ = __email__ = ''

# extract version
#with open(os.path.join(os.path.dirname(__file__), "episcanpy", "version.py")) as f:
#    version = f.read().split("\n")[0].split("=")[-1].strip(' ').strip('"')

setup(
    name='episcanpy',
    version=versioneer.get_version(),
    #version=version,
    cmdclass=versioneer.get_cmdclass(),
    description='Epigenomics Single-Cell Analysis in Python.',
    long_description=Path('README.rst').read_text('utf-8'),
    #long_description=read_file('README.rst'),
    url='http://github.com/colomemaria/epiScanpy',
    #download_url='https://github.com/colomemaria/epiScanpy/archive/v0.1-beta.tar.gz',
    author=__author__,
    author_email=__email__,
    license='BSD',
    python_requires='>=3.5',
    install_requires=[
        l.strip() for l in
        Path('requirements.txt').read_text('utf-8').splitlines()
    ],
    extras_require=dict(
        louvain=['python-igraph', 'louvain>=0.6'],
        leiden=['python-igraph', 'leidenalg'],
        bbknn=['bbknn'],
        combat=['patsy'],
        doc=['sphinx', 'sphinx_rtd_theme', 'sphinx_autodoc_typehints', 'scanpydoc'],
        test=['pytest'],
    ),
    packages=find_packages(),
    # `package_data` does NOT work for source distributions!!!
    # you also need MANIFTEST.in
    # https://stackoverflow.com/questions/7522250/how-to-include-package-data-with-setuptools-distribute
    package_data={'': '*.txt'},
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Framework :: Jupyter',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
)



import numpy as np
from numpy import NaN

# chromosomes for 2 principal species. If you work with another genome
# the chromosomes will have to be specified
HUMAN = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '2',
         '20', '21', '22', '3', '4', '5', '6', '7', '8', '9', 'X', 'Y']
MOUSE = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '2',
         '3', '4', '5', '6', '7', '8', '9', 'X', 'Y']
# my favorite dataset ;-)
HEAD_Ecker = ['chr', 'pos', 'strand', 'mc_class', 'mc_count', 'total', 'methylated\n']



def methylation_level(reduced_cyt, feature, chromosome, threshold=1):
    """
    Measure the methylation for the feature you give as input using the reduce
    representation of the sample cytosine (output of read_methylation_file)
    Parameters
    ----------
    reduced_cyt: datatype that contained processed sample file. It only contains
        the cytosines that were in the genomic context you wanted to filter for.
        (output of read_methylation_file function).
    feature: the feature in the right datatype for which you want to determine
        the methylation level.
    chromosome: chromosomes if the species you are considering. default value
        is the human genome (including mitochondrial and sexual chromosomes).
    """
    ## actually, to write sparse matrix I need a list, not a dictionary
    #meth_levels_bins = {key:[] for key in chromosome}
    meth_levels_bins = []

    for c in chromosome:
        meth_reads = np.zeros(len(feature[c]))
        tot_reads = np.zeros(len(feature[c]))
        nb_cyt = np.zeros(len(feature[c]))
        cytosines = reduced_cyt[c] 
        i = 0
        for j in range(len(feature[c])): # for every bins in a given chrom
            meth_reads = 0.0
            tot_reads = 1.0
            nb_cyt = 0
            # I am skipping cytosine that are before the beginning 
            # of my current bin. 
            while (i < len(cytosines)) and cytosines[i][0] < feature[c][j][0]:
                i += 1
            # Once I got one cytosine that is after the beginning of my feature 
            # I need to check if this feature is within the enhancer limits
            # so if the position of the cytosine is not > to the end of the feature
            if i<len(cytosines) and cytosines[i][0] <= feature[c][j][1]:
                meth_reads += cytosines[i][-2] # meth cyt read
                tot_reads += cytosines[i][-1] # tot cyt read
                nb_cyt += 1 # nb of cyt

            # check if the next cytosine fall into the current feature is important
            # to this end, I have another pointer/iterator k. 
            # at the next feature I will have to check from i but for the current 
            # feature I need to check the next cytosine and I use the variable k for 
            # this.
            k = i+1
            while k < len(cytosines) and cytosines[k][0] <= feature[c][j][1]:
                meth_reads += cytosines[i][-2] # meth cyt read
                tot_reads += cytosines[i][-1] # tot cyt read
                nb_cyt += 1  # nb of cyt
                k += 1
            ## actually, to write sparse matrix I need a list, not a dictionary
            if nb_cyt >= threshold:
                #meth_levels_bins[c].append(format(meth_reads/tot_reads, '.3f'))
                meth_levels_bins.append(format(meth_reads/tot_reads, '.3f'))
            else:
                #meth_levels_bins[c].append(np.nan)
                meth_levels_bins.append(np.nan)


    return(meth_levels_bins)

def write_not_sparse_meth(meth_to_write, name_file, writing_option='a', feature_names=None, cell_names=None):
    """
    takes the output of prep_methlevels and write everything in a given file.
    
    Parameters
    ----------
    """
    with open(name_file, writing_option) as file_to_write:
        if writing_option == 'w':
            file_to_write.write('%sparse matrix of methylation levels +1\n')
            file_to_write.write('%missing values are NaN not zeros.\n')
            if cell_names != None:
                file_to_write.write('%cell_names\t')
                file_to_write.write('\t'.join(cell_names))
                file_to_write.write('\n')
            if feature_names != None:
                file_to_write.write("feature_names\t")
                file_to_write.write('\t'.join(feature_names))
                file_to_write.write('\n')
    
        file_to_write.write('\t'.join([str(value) for value in meth_to_write]))
        file_to_write.write('\n')



def extract_methylation(sample_name, feature, meth_type=None, path='', #head=HEAD_Ecker,
	chromosome=HUMAN, threshold=1, col=None, write=False, writing_option='a', cell_names=None):
    """
    This function read the sample to produce a reduced representation of the
    sample (including only cytosines in a certain context) and then get
    corresponding methylation level in the feature given as input.
    Parameters
    ----------
    sample_name:
        name of the file to read to extract key information.
    feature:
    meth_type:
        CG, CH or not specified
    head: if there is header that you don't want to read. An annotation in the
        file you are reading. The default value is the Methylpy/Ecker header
    path: path of the to access the file of the sample you want to read. 
    chromosome: chromosomes if the species you are considering. default value
        is the human genome (including mitochondrial and sexual chromosomes)
    cell_names is the list of cells you want to put as annotation at the beginning of your file
    """
    # It correspond to the annotation in the methylation call for methylpy.
    # (at least for the Ecker dataset)
    # I specify it here so I can automatically skip the line when I am reading the
    # methylation files
    ### doesn't work yet. You must have met reads and tot reads. 
    ## col argument contain the columns. We expect a list with the columns like such:
    # col = [chrom, pos_cyt, met_reads, unmet_reads, tot_reads]
    #If there is a missing column (either met_reads, unmer_reads, tot_reads) put 'NA'
    # also weird header ??
    if col != None:
    	chrom=col[0]
    	pos=col[1]
    	tot = col[3]
    	status=col[4]
    else:
    	chrom=0
    	pos=1
    	met=4
    	tot=5
    	status=3


    if meth_type == None:
    	reduced_cyt = read_meth_file(sample_name, path, chromosome, pos, met, tot, status)
    elif meth_type == 'CG':
    	reduced_cyt = read_meth_fileCG(sample_name, path, chromosome, pos, met, tot, status)
    elif meth_type == 'CH':
    	reduced_cyt = read_meth_fileCH(sample_name, path, chromosome, pos, met, tot, status)
    else:
    	# print a warning saying that the argument is not valid. We take all cytosines
    	reduced_cyt = read_meth_file(sample_name, head, path, chromosome)

    final_output = methylation_level(reduced_cyt, feature, chromosome, threshold)
    if write:
    	# I need to extract feature_names:
    	feature_names = extract_feature_names(feature)
    	write_not_sparse_meth(final_output, sample_name, writing_option, feature_names, cell_names)
    return(final_output)

## When I am making the feature names while generating features I should add the chromosome info.
def extract_feature_names(feature):
    bin_names = []
    for c in sorted(feature.keys()):
        bin_names += [x[-1] for x in feature[c]]

    return(bin_names)

def read_meth_fileCG(sample_name, path, chromosome, chrom=0, pos=1, met=4, tot=5, status=3):
    """
    Read file from which you want to extract the methylation level and
    (assuming it is like the Ecker/Methylpy format) extract the number of
    methylated read and the total number of read for the cytosines covered and
    in the right genomic context (CG or CH)
    Parameters
    ----------
    sample_name:
        name of the file to read to extract key information.
    meth_type:
        CG, CH or not specified
    head: if there is header that you don't want to read. An annotation in the
        file you are reading. The default value is the Methylpy/Ecker header
    path: path of the to access the file of the sample you want to read. 
    chromosome: chromosomes if the species you are considering. default value
        is the human genome (including mitochondrial and sexual chromosomes)
    """
    reduced_cyt = {key: [] for key in chromosome} # to store cyt for each chrom (intermed output)
    with open(path+sample_name) as sample:
        for line in sample:
            line = line.split('\t')
            if (line[status] in ['CGG', 'CGC', 'CGA', 'CGT']):
                reduced_cyt[line[chrom]].append((int(line[pos]), int(line[met]), int(line[tot])))
    return(reduced_cyt)

def read_meth_fileCH(sample_name, path, chromosome, chrom=0, pos=1, met=4, unmet=5, status=3):
    """
    Read file from which you want to extract the methylation level and
    (assuming it is like the Ecker/Methylpy format) extract the number of
    methylated read and the total number of read for the cytosines covered and
    in the right genomic context (CG or CH)
    Parameters
    ----------
    sample_name:
        name of the file to read to extract key information.
    meth_type:
        CG, CH or not specified
    head: if there is header that you don't want to read. An annotation in the
        file you are reading. The default value is the Methylpy/Ecker header
    path: path of the to access the file of the sample you want to read. 
    chromosome: chromosomes if the species you are considering. default value
        is the human genome (including mitochondrial and sexual chromosomes)
    """
    reduced_cyt = {key: [] for key in chromosome} # to store cyt for each chrom (intermed output)
    with open(path+sample_name) as sample:
        for line in sample:
            line = line.split('\t')
            if (line[status] not in ['CGG', 'CGC', 'CGA', 'CGT', 'mc_class']):
                reduced_cyt[line[chrom]].append((int(line[pos]), int(line[met]), int(line[unmet])))
    return(reduced_cyt)

def read_meth_file(sample_name, path, chromosome, chrom=0, pos=1, met=4, unmet=5, status=3):
    """
    Read file from which you want to extract the methylation level and
    (assuming it is like the Ecker/Methylpy format) extract the number of
    methylated read and the total number of read for the cytosines covered and
    in the right genomic context (CG or CH)
    Parameters
    ----------
    sample_name:
        name of the file to read to extract key information.
    meth_type:
        CG, CH or not specified
    head: if there is header that you don't want to read. An annotation in the
        file you are reading. The default value is the Methylpy/Ecker header
    path: path of the to access the file of the sample you want to read. 
    chromosome: chromosomes if the species you are considering. default value
        is the human genome (including mitochondrial and sexual chromosomes)
    """
    reduced_cyt = {key: [] for key in chromosome} # to store cyt for each chrom (intermed output)
    with open(path+sample_name) as sample:
        for line in sample:
            line = line.split('\t')
            if line[status] != 'mc_class':
                reduced_cyt[line[chrom]].append((int(line[pos]), int(line[met]), int(line[unmet])))
    return(reduced_cyt)
     
def filter_and_average_features(methylation_feature, feature, cell_name = '', split=False, threshold=1, average=''):
    """
    This function can do 3 different things :
        1- It is possible to filter features based on the number of cytosine covered per unit. 
        2- And/or to split the different feature contained. For example, a bed file that contain all
        TFBS can be consider as one type of feature or it can be split by the different TF.
        3- it can average the methylation level as a general average over all unit covered
        or it can average the average methylation level at each unit covered.
        
    Parameters
    ----------
    methylation_feature:    the data type that contain methylation level for the
        feature considered
    feature:    the feature data type for which you have the methylation level
    split:      do you want to split based on the feature names
    threshold:  the minimum number of cytosine per feature to keep it 
    average:    the type of average you want ('general', 'average', anything else mean
    no average done)
    """
    final_output = [cell_name]
    list_methylation = make_list(methylation_feature)
    if split==True:
        list_feature = [x[-1] for x in make_list(feature)] 
    else:
        list_feature = ['standard_name' for x in range(len(list_methylation))]
    name_feature = list(set(list_feature))
    output_average = {}
    
    if average == 'general':
        for keys in name_feature:
            output_average[keys] = [0,0]
        for index in range(len(list_methylation)):
            if list_methylation[index][-1] >= threshold:
                output_average[list_feature[index]][0] += list_methylation[index][0]
                output_average[list_feature[index]][1] += list_methylation[index][1]

    elif average == 'average':
        for keys in name_feature:
            output_average[keys] = []
        for index in range(len(list_methylation)):
            if list_methylation[index][-1] >= threshold and list_methylation[index][1]>0:
                output_average[list_feature[index]].append(list_methylation[index][0]/list_methylation[index][1])
        #return(output_average)
    # can be optimized by doing 2 arrays and then divide every elements
                
    else:
        for keys in name_feature:
            output_average[keys] = []
        for index in range(len(list_methylation)):
            if list_methylation[index][-1] >= threshold:
                output_average[list_feature[index]].append(list_methylation[index][0:2])
        return(output_average)
        
    for key in sorted(output_average.keys()):
        if average == 'average':
            final_output.append(np.average(output_average[key]))
        elif average == 'general' and output_average[key][1]>0:
            final_output.append(float(output_average[key][0])/output_average[key][1])
        else:
            final_output.append("nan")
        
    return(final_output)
    
def filter_and_average_features_chrm(methylation_feature, feature, cell_name = '', split=False, threshold=1, average=''):
    """
    This function can do 3 different things :
        1- It is possible to filter features based on the number of cytosine covered per unit. 
        2- And/or to split the different feature contained. For example, a bed file that contain all
        TFBS can be consider as one type of feature or it can be split by the different TF.
        3- it can average the methylation level as a general average over all unit covered
        or it can average the average methylation level at each unit covered.   
    Parameters
    ----------
    methylation_feature:    the data type that contain methylation level for the
        feature considered
    feature:    the feature data type for which you have the methylation level
    split:      do you want to split based on the feature names
    threshold:  the minimum number of cytosine per feature to keep it 
    average:    the type of average you want ('general', 'average', anything else mean
    no average done)
    """
    final_output = [cell_name]
    for chrm in methylation_feature.keys():
        list_methylation = [x for x in methylation_feature[chrm]]
        if split==True:
        	list_feature = [x[-1] for x in make_list(feature)] 
        else:
        	list_feature = ['standard_name' for x in range(len(list_methylation))]
        name_feature = list(set(list_feature))
        output_average = {}
        if average == 'general':
            for keys in name_feature:
                output_average[keys] = [0,0]
            for index in range(len(list_methylation)):
                if list_methylation[index][-1] >= threshold:
                    output_average[list_feature[index]][0] += list_methylation[index][0]
                    output_average[list_feature[index]][1] += list_methylation[index][1]

        elif average == 'average':
            for keys in name_feature:
                output_average[keys] = []
            for index in range(len(list_methylation)):
                if list_methylation[index][-1] >= threshold  and list_methylation[index][1]>0:
                    output_average[list_feature[index]].append(list_methylation[index][0]/list_methylation[index][1])
        #return(output_average)
    # can be optimized by doing 2 arrays and then divide every elements
        else:
            for keys in name_feature:
                output_average[keys] = []
            for index in range(len(list_methylation)):
                if list_methylation[index][-1] >= threshold:
                    output_average[list_feature[index]].append(list_methylation[index][0:2])
            return(output_average)

        for key in sorted(output_average.keys()):
            if average == 'average':
                final_output.append(np.average(output_average[key]))
            elif average == 'general' and output_average[key][1]>0:
                final_output.append(float(output_average[key][0])/output_average[key][1])
            else:
                final_output.append("nan")

    return(final_output)

def prep_methlevels(methylation_feature, name_cell, threshold=1):
    """
    Takes methylation levels, replace bins that are not covered enough by NaN,
    otherwise get the average methylation level for the bin and return the final
    values in alist. Ready to write. 
    
    Parameters
    ----------
    """
    output_list = [name_cell]
    for element in make_list(methylation_feature):
        if element[2] >= threshold:
            output_list.append(float(element[0])/element[1])
        else:
            output_list.append(NaN)
    return(output_list)
  
def write_list(the_list, file_to_write):
    for value in the_list:
        file_to_write.write(str(value))
        file_to_write.write("\t")
    file_to_write.write("\n")
    return()
    
def write_methlevel(meth_to_write, name_file, cell, writing_option='a', feature_names=None):
    """
    takes the output of prep_methlevels and write everything in a given file.
    
    Parameters
    ----------
    """
    file_to_write = open(name_file, writing_option)

    if feature_names != None:
        feature_to_write = ["sample_name"]
        feature_to_write += sorted(feature_names)
        write_list(feature_to_write, file_to_write) # should add an condition
                                            # to be sure features is a dictionary
    meth_to_write = [cell] + meth_to_write
    write_list(meth_to_write, file_to_write)
    file_to_write.close()
    return()

def make_list(dico):
    """
    Make information stored in  dictionary with chromosomes as key
    into one list. (lost of chromosomal information)
    """
    final_list = []
    for c in sorted(dico.keys()):
        final_list += dico[c]
    return final_list

def build_count_mtx(cells, annotation, path="", output_file=None, writing_option="a",
                    meth_context="CG", chromosome=MOUSE, feature_names=None,
                   threshold=1, ct_mtx=None, sparse=False):
    """
    Build methylation count matrix for a given annotation.
    It either write the count matrix (if given an output file) or return it as a variable (numpy matrix). 
    If you want to add cells to an already existing matrix (with the same annotations),  you put the initial matrix as ct_mtx or you specify the matrix to write +
    writing option = a
        
    if you want to write down the matrix as a sparse matrix you have to specify it (not implented yet)
        
    I need to pay attention to where I am writing the output file.
        
    Also, verbosity..
        
    Pay attention, it does not average variables. If you want to process many small features such as
    tfbs, we advise to use the dedicated function.
        
    Parameters
    ----------
    cells:
        list of the file names to read to build the count matrix.
    annotation:
        loaded annotation to use to build the count matrix
        'str' or 'list' depending of the number of matrices to build
    path:
        path to the input data. 
    output_file:
        name files to write. 'str' or 'list' depending of the number of matrices to build
    writing_option:
        either 'w' if you want to erase potentialy already existing file or 'a' to append.
        'str' or 'list' if you have a list of matrices and the writing options are differents
    meth_context:
        read methylation in 'CG' of 'CH' context
    chromosome:
        'MOUSE' and 'HUMAN' (without mitochondrial genome) or list with chromosomes.
    feature_names:
        If you want to write down the name of the annotation features. 
        'Int' (or 'list' if you have multiple annotations)
    thereshold:
        the minimum of cytosines covered per annotation to calculate a methylation level.
        default=1 'Int' (or 'list' if you have multiple annotations with different thresholds)
    ct_mtx:
        numpy matrix containing the same set of annotations and for which you want to append.
        default: None
    sparse:
        Boolean, writing option as a normal or sparse matrix.
        default: False
    
    """
    #verbosity
    i = 0
    
    #################################
    if type(annotation) != list:
        annotation = [annotation]
        output_file = [output_file]
        ct_mtx = [ct_mtx]
        feature_names = [feature_names]
        
    nb_annotation = len(annotation)
    
    if type(writing_option) != list:
        writing_option = [writing_option for x in range(nb_annotation)]
    if type(threshold) != list:
        threshold = [threshold for x in range(nb_annotation)]
    if (output_file != None):
        if (type(output_file) != list):
            output_file = [output_file]
        
    #################################
    for cell in cells:
        #verbosity
        print(i, cell)
        i += 1
        
        # read the file to extract cytosines in the right context
        if meth_context == 'CG':
            tmp_file = read_meth_fileCG(cell, path, chromosome)
        elif meth_context == 'CH':
            tmp_file = read_meth_fileCH(cell, path, chromosome)
        else:
            break
        
        # build the cell vector for the count matrix at every set of annotations
        for index_annot in range(nb_annotation):
            meth_level_annot = methylation_level(tmp_file, annotation[index_annot], chromosome, threshold[index_annot])
            if type(output_file) == list:
                write_methlevel(meth_level_annot, output_file[index_annot], cell, writing_option[index_annot], feature_names[index_annot])
            else:
                if ct_mtx == None:
                    ct_mtx = [np.matrix(meth_level_annot)]
                    #ct_mtx[index_annot] = np.matrix(meth_level_annot)
                elif index_annot>=len(ct_mtx):
                    ct_mtx.append(np.matrix(meth_level_annot))
                else:
                    ct_mtx[index_annot] = np.vstack([ct_mtx[index_annot], meth_level_annot])
                
    if ct_mtx != None:
        return(ct_mtx)
    else:
        return()



def extract_CG(input_sample, output_variable, head):
    """
    Description function. 
    
    Extract relevant information for cytosines covered in a CG context
    It supposes that it is on the same format at the Methylpy/Ecker data
    Parameters
    ----------
    input_sample: file opened that contain the cytosines
    output_variable: the datatype that gather the methylation levels for a
        given type of features
    head: annotation of the file that we are reading (line to skip)
    """
    for cyt_called in input_sample:
        cyt_called = cyt_called.split("\t")
        if (cyt_called != head) and (cyt_called[3][0:2] == 'CG'):
            output_variable[cyt_called[0]].append([int(cyt_called[1]),
                       int(cyt_called[-3]), int(cyt_called[-2])])
    return(output_variable)
    
def extract_CH(sample, output_variable, head):
    """
    Description function. 
    
    Extract relevant information for cytosines covered in a CH context
    It supposes that it is on the same format at the Methylpy/Ecker data
    Parameters
    ----------
    input_sample: file opened that contain the cytosines
    output_variable: the datatype that gather the methylation levels for a
        given type of features
    head: annotation of the file that we are reading (line to skip)
    """
    for cyt_called in sample:
        cyt_called = cyt_called.split("\t")
        if cyt_called != head and cyt_called[3][0:2] != 'CG':
            output_variable[cyt_called[0]].append([int(cyt_called[1]),
                       int(cyt_called[-3]), int(cyt_called[-2])])
    return(output_variable)



"""Utility functions and classes
"""

import sys
import inspect
from collections import namedtuple
from functools import partial
from types import ModuleType
from typing import Union, Callable, Optional

import numpy as np
import scipy.sparse
from natsort import natsorted
from textwrap import dedent
from pandas.api.types import CategoricalDtype

from . import settings
from . import logging as logg

EPS = 1e-15


def check_versions():
    import warnings
    from distutils.version import LooseVersion

    if sys.version_info < (3, 0):
        warnings.warn('Scanpy only runs reliably with Python 3, preferrably >=3.5.')

    import anndata
    # NOTE: pytest does not correctly retrieve anndata's version? why?
    #       use the following hack...
    if anndata.__version__ != '0+unknown':
        if anndata.__version__ < LooseVersion('0.6.10'):
            raise ImportError('Scanpy {} needs anndata version >=0.6.10, not {}.\n'
                              'Run `pip install anndata -U --no-deps`.'
                              .format(__version__, anndata.__version__))



def getdoc(c_or_f: Union[Callable, type]) -> Optional[str]:
    if getattr(c_or_f, '__doc__', None) is None:
        return None
    doc = inspect.getdoc(c_or_f)
    if isinstance(c_or_f, type) and hasattr(c_or_f, '__init__'):
        sig = inspect.signature(c_or_f.__init__)
    else:
        sig = inspect.signature(c_or_f)

    def type_doc(name: str):
        param = sig.parameters[name]  # type: inspect.Parameter
        cls = getattr(param.annotation, '__qualname__', repr(param.annotation))
        if param.default is not param.empty:
            return '{}, optional (default: {!r})'.format(cls, param.default)
        else:
            return cls

    return '\n'.join(
        '{} : {}'.format(line, type_doc(line)) if line.strip() in sig.parameters else line
        for line in doc.split('\n')
    )


def descend_classes_and_funcs(mod: ModuleType, root: str):
    for obj in vars(mod).values():
        if not getattr(obj, '__module__', getattr(obj, '__qualname__', getattr(obj, '__name__', ''))).startswith(root):
            continue
        if isinstance(obj, Callable):
            yield obj
            if isinstance(obj, type):
                yield from (m for m in vars(obj).values() if isinstance(m, Callable))
        elif isinstance(obj, ModuleType):
            yield from descend_classes_and_funcs(obj, root)


def annotate_doc_types(mod: ModuleType, root: str):
    for c_or_f in descend_classes_and_funcs(mod, root):
        c_or_f.getdoc = partial(getdoc, c_or_f)


def doc_params(**kwds):
    """\
    Docstrings should start with "\" in the first line for proper formatting.
    """
    def dec(obj):
        obj.__doc__ = dedent(obj.__doc__).format(**kwds)
        return obj
    return dec


def merge_groups(adata, key, map_groups, key_added=None, map_colors=None):
    """
    Parameters
    ----------
    map_colors : `dict`
        Dict with color specification for new groups that have no corresponding
        old group.
    """
    if key_added is None:
        key_added = key + '_merged'
    adata.obs[key_added] = adata.obs[key].map(
        map_groups).astype(CategoricalDtype())
    old_categories = adata.obs[key].cat.categories
    new_categories = adata.obs[key_added].cat.categories
    # map_colors is passed
    if map_colors is not None:
        old_colors = None
        if key + '_colors' in adata.uns:
            old_colors = adata.uns[key + '_colors']
        new_colors = []
        for group in adata.obs[key_added].cat.categories:
            if group in map_colors:
                new_colors.append(map_colors[group])
            elif group in old_categories and old_colors is not None:
                new_colors.append(old_colors[old_categories.get_loc(group)])
            else:
                raise ValueError('You didn\'t specify a color for {}.'
                                 .format(group))
        adata.uns[key_added + '_colors'] = new_colors
    # map_colors is not passed
    elif key + '_colors' in adata.uns:
        old_colors = adata.uns[key + '_colors']
        inverse_map_groups = {g: [] for g in new_categories}
        for old_group in old_categories:
            inverse_map_groups[map_groups[old_group]].append(old_group)
        new_colors = []
        for group in new_categories:
            # take the largest of the old groups
            old_group = adata.obs[key][adata.obs[key].isin(
                inverse_map_groups[group])].value_counts().index[0]
            new_colors.append(old_colors[old_categories.get_loc(old_group)])
        adata.uns[key_added + '_colors'] = new_colors


# --------------------------------------------------------------------------------
# Graph stuff
# --------------------------------------------------------------------------------


def cross_entropy_neighbors_in_rep(adata, use_rep, n_points=3):
    """Compare neighborhood graph of representation based on cross entropy.

    `n_points` denotes the number of points to add as highlight annotation.

    Returns
    -------
    The cross entropy and the geodesic-distance-weighted cross entropy as
    ``entropy, geo_entropy_d, geo_entropy_o``.

    Adds the most overlapping or disconnected points as annotation to `adata`.
    """
    # see below why we need this
    if 'X_diffmap' not in adata.obsm.keys():
        raise ValueError('Run `tl.diffmap` on `adata`, first.')

    adata_ref = adata  # simple renaming, don't need copy here
    adata_cmp = adata.copy()
    n_neighbors = adata_ref.uns['neighbors']['params']['n_neighbors']
    from .neighbors import neighbors
    neighbors(adata_cmp, n_neighbors=n_neighbors, use_rep=use_rep)
    from .tools.diffmap import diffmap
    diffmap(adata_cmp)

    graph_ref = adata_ref.uns['neighbors']['connectivities']
    graph_cmp = adata_cmp.uns['neighbors']['connectivities']

    graph_ref = graph_ref.tocoo()  # makes a copy
    graph_cmp = graph_cmp.tocoo()

    edgeset_ref = {e for e in zip(graph_ref.row, graph_ref.col)}
    edgeset_cmp = {e for e in zip(graph_cmp.row, graph_cmp.col)}
    edgeset_union = list(edgeset_ref.union(edgeset_cmp))

    edgeset_union_indices = tuple(zip(*edgeset_union))
    edgeset_union_indices = (np.array(edgeset_union_indices[0]), np.array(edgeset_union_indices[1]))

    n_edges_ref = len(graph_ref.nonzero()[0])
    n_edges_cmp = len(graph_cmp.nonzero()[0])
    n_edges_union = len(edgeset_union)
    logg.msg(
        '... n_edges_ref', n_edges_ref,
        'n_edges_cmp', n_edges_cmp,
        'n_edges_union', n_edges_union)

    graph_ref = graph_ref.tocsr()  # need a copy of the csr graph anyways
    graph_cmp = graph_cmp.tocsr()

    p_ref = graph_ref[edgeset_union_indices].A1
    p_cmp = graph_cmp[edgeset_union_indices].A1

    # the following is how one compares it to log_loss form sklearn
    # p_ref[p_ref.nonzero()] = 1
    # from sklearn.metrics import log_loss
    # print(log_loss(p_ref, p_cmp))
    p_cmp = np.clip(p_cmp, EPS, 1-EPS)
    ratio = np.clip(p_ref / p_cmp, EPS, None)
    ratio_1m = np.clip((1 - p_ref) / (1 - p_cmp), EPS, None)

    entropy = np.sum(p_ref * np.log(ratio) + (1-p_ref) * np.log(ratio_1m))

    n_edges_fully_connected = (graph_ref.shape[0]**2 - graph_ref.shape[0])
    entropy /= n_edges_fully_connected

    fraction_edges = n_edges_ref / n_edges_fully_connected
    naive_entropy = (fraction_edges * np.log(1./fraction_edges)
                     + (1-fraction_edges) * np.log(1./(1-fraction_edges)))
    logg.msg('cross entropy of naive sparse prediction {:.3e}'.format(naive_entropy))
    logg.msg('cross entropy of random prediction {:.3e}'.format(-np.log(0.5)))
    logg.info('cross entropy {:.3e}'.format(entropy))

    # for manifold analysis, restrict to largest connected component in
    # reference
    # now that we clip at a quite high value below, this might not even be
    # necessary
    n_components, labels = scipy.sparse.csgraph.connected_components(graph_ref)
    largest_component = np.arange(graph_ref.shape[0], dtype=int)
    if n_components > 1:
        component_sizes = np.bincount(labels)
        logg.msg('largest component has size', component_sizes.max())
        largest_component = np.where(
            component_sizes == component_sizes.max())[0][0]
        graph_ref_red = graph_ref.tocsr()[labels == largest_component, :]
        graph_ref_red = graph_ref_red.tocsc()[:, labels == largest_component]
        graph_ref_red = graph_ref_red.tocoo()
        graph_cmp_red = graph_cmp.tocsr()[labels == largest_component, :]
        graph_cmp_red = graph_cmp_red.tocsc()[:, labels == largest_component]
        graph_cmp_red = graph_cmp_red.tocoo()
        edgeset_ref_red = {e for e in zip(graph_ref_red.row, graph_ref_red.col)}
        edgeset_cmp_red = {e for e in zip(graph_cmp_red.row, graph_cmp_red.col)}
        edgeset_union_red = edgeset_ref_red.union(edgeset_cmp_red)
        map_indices = np.where(labels == largest_component)[0]
        edgeset_union_red = {
            (map_indices[i], map_indices[j]) for (i, j) in edgeset_union_red}

    from .neighbors import Neighbors
    neigh_ref = Neighbors(adata_ref)
    dist_ref = neigh_ref.distances_dpt  # we expect 'X_diffmap' to be already present

    neigh_cmp = Neighbors(adata_cmp)
    dist_cmp = neigh_cmp.distances_dpt

    d_cmp = np.zeros_like(p_ref)
    d_ref = np.zeros_like(p_ref)
    for i, e in enumerate(edgeset_union):
        # skip contributions that are not in the largest component
        if n_components > 1 and e not in edgeset_union_red:
            continue
        d_cmp[i] = dist_cmp[e]
        d_ref[i] = dist_ref[e]

    MAX_DIST = 1000
    d_cmp = np.clip(d_cmp, 0.1, MAX_DIST)  # we don't want to measure collapsing clusters
    d_ref = np.clip(d_ref, 0.1, MAX_DIST)

    weights = np.array(d_cmp / d_ref)            # disconnected regions
    weights_overlap = np.array(d_ref / d_cmp)    # overlapping regions

    # the following is just for annotation of figures
    if 'highlights' not in adata_ref.uns:
        adata_ref.uns['highlights'] = {}
    else:
        # remove old disconnected and overlapping points
        new_highlights = {}
        for k, v in adata_ref.uns['highlights'].items():
            if v != 'O' and v not in {'D0', 'D1', 'D2', 'D3', 'D4'}:
                new_highlights[k] = v
        adata_ref.uns['highlights'] = new_highlights

    # points that are maximally disconnected
    max_weights = np.argpartition(weights, kth=-n_points)[-n_points:]
    points = list(edgeset_union_indices[0][max_weights])
    points2 = list(edgeset_union_indices[1][max_weights])
    found_disconnected_points = False
    for ip, p in enumerate(points):
        if d_cmp[max_weights][ip] == MAX_DIST:
            adata_ref.uns['highlights'][p] = 'D' + str(ip)
            adata_ref.uns['highlights'][points2[ip]] = 'D' + str(ip)
            found_disconnected_points = True
    if found_disconnected_points:
        logg.msg('most disconnected points', points)
        logg.msg('    with weights', weights[max_weights].round(1))

    max_weights = np.argpartition(
        weights_overlap, kth=-n_points)[-n_points:]
    points = list(edgeset_union_indices[0][max_weights])
    for p in points:
        adata_ref.uns['highlights'][p] = 'O'
    logg.msg('most overlapping points', points)
    logg.msg('    with weights', weights_overlap[max_weights].round(1))
    logg.msg('    with d_rep', d_cmp[max_weights].round(1))
    logg.msg('    with d_ref', d_ref[max_weights].round(1))

    geo_entropy_d = np.sum(weights * p_ref * np.log(ratio))
    geo_entropy_o = np.sum(weights_overlap * (1-p_ref) * np.log(ratio_1m))

    geo_entropy_d /= n_edges_fully_connected
    geo_entropy_o /= n_edges_fully_connected

    logg.info('geodesic cross entropy {:.3e}'.format(geo_entropy_d + geo_entropy_o))
    return entropy, geo_entropy_d, geo_entropy_o


def get_graph_tool_from_adjacency(adjacency, directed=None):
    """Get graph_tool graph from adjacency matrix."""
    import graph_tool as gt
    adjacency_edge_list = adjacency
    if not directed:
        from scipy.sparse import tril
        adjacency_edge_list = tril(adjacency)
    g = gt.Graph(directed=directed)
    g.add_vertex(adjacency.shape[0])  # this adds adjacency.shap[0] vertices
    g.add_edge_list(np.transpose(adjacency_edge_list.nonzero()))
    weights = g.new_edge_property('double')
    for e in g.edges():
        # graph_tool uses the following convention,
        # which is opposite to the rest of scanpy
        weights[e] = adjacency[int(e.source()), int(e.target())]
    g.edge_properties['weight'] = weights
    return g


def get_igraph_from_adjacency(adjacency, directed=None):
    """Get igraph graph from adjacency matrix."""
    import igraph as ig
    sources, targets = adjacency.nonzero()
    weights = adjacency[sources, targets]
    if isinstance(weights, np.matrix):
        weights = weights.A1
    g = ig.Graph(directed=directed)
    g.add_vertices(adjacency.shape[0])  # this adds adjacency.shap[0] vertices
    g.add_edges(list(zip(sources, targets)))
    try:
        g.es['weight'] = weights
    except:
        pass
    if g.vcount() != adjacency.shape[0]:
        logg.warn('The constructed graph has only {} nodes. '
                  'Your adjacency matrix contained redundant nodes.'
                  .format(g.vcount()))
    return g


def get_sparse_from_igraph(graph, weight_attr=None):
    from scipy.sparse import csr_matrix
    edges = graph.get_edgelist()
    if weight_attr is None:
        weights = [1] * len(edges)
    else:
        weights = graph.es[weight_attr]
    if not graph.is_directed():
        edges.extend([(v, u) for u, v in edges])
        weights.extend(weights)
    shape = graph.vcount()
    shape = (shape, shape)
    if len(edges) > 0:
        return csr_matrix((weights, zip(*edges)), shape=shape)
    else:
        return csr_matrix(shape)


def compute_association_matrix_of_groups(adata, prediction, reference,
                                         normalization='prediction',
                                         threshold=0.01, max_n_names=2):
    """Compute overlaps between groups.

    See ``identify_groups`` for identifying the groups.

    Parameters
    ----------
    adata : AnnData
    prediction : str
        Field name of adata.obs.
    reference : str
        Field name of adata.obs.
    normalization : {'prediction', 'reference'}
        Whether to normalize with respect to the predicted groups or the
        reference groups.
    threshold : float, optional (default: 0.01)
        Do not consider associations whose overlap is below this fraction.
    max_n_names : int or None, optional (default: 2)
        Control how many reference names you want to be associated with per
        predicted name. Set to `None`, if you want all.

    Returns
    -------
    Tuple of
    asso_names : list of associated reference names (`max_n_names` for each
        predicted name)
    asso_matrix : matrix where rows correspond to the predicted labels and
        columns to the reference labels, entries are proportional to degree of
        association
    """
    if normalization not in {'prediction', 'reference'}:
        raise ValueError('`normalization` needs to be either "prediction" or "reference".')
    sanitize_anndata(adata)
    cats = adata.obs[reference].cat.categories
    for cat in cats:
        if cat in settings.categories_to_ignore:
            logg.info('Ignoring category \'{}\' '
                      'as it\'s in `settings.categories_to_ignore`.'
                      .format(cat))
    asso_names = []
    asso_matrix = []
    for ipred_group, pred_group in enumerate(
            adata.obs[prediction].cat.categories):
        if '?' in pred_group: pred_group = str(ipred_group)
        # starting from numpy version 1.13, subtractions of boolean arrays are deprecated
        mask_pred = adata.obs[prediction].values == pred_group
        mask_pred_int = mask_pred.astype(np.int8)
        asso_matrix += [[]]
        for ref_group in adata.obs[reference].cat.categories:
            mask_ref = (adata.obs[reference].values == ref_group).astype(np.int8)
            mask_ref_or_pred = mask_ref.copy()
            mask_ref_or_pred[mask_pred] = 1
            # e.g. if the pred group is contained in mask_ref, mask_ref and
            # mask_ref_or_pred are the same
            if normalization == 'prediction':
                # compute which fraction of the predicted group is contained in
                # the ref group
                ratio_contained = (np.sum(mask_pred_int) -
                    np.sum(mask_ref_or_pred - mask_ref)) / np.sum(mask_pred_int)
            else:
                # compute which fraction of the reference group is contained in
                # the predicted group
                ratio_contained = (np.sum(mask_ref) -
                    np.sum(mask_ref_or_pred - mask_pred_int)) / np.sum(mask_ref)
            asso_matrix[-1] += [ratio_contained]
        name_list_pred = [cats[i] if cats[i] not in settings.categories_to_ignore else ''
                          for i in np.argsort(asso_matrix[-1])[::-1]
                          if asso_matrix[-1][i] > threshold]
        asso_names += ['\n'.join(name_list_pred[:max_n_names])]
    Result = namedtuple('compute_association_matrix_of_groups',
                        ['asso_names', 'asso_matrix'])
    return Result(asso_names=asso_names, asso_matrix=np.array(asso_matrix))


def get_associated_colors_of_groups(reference_colors, asso_matrix):
    asso_colors = [{reference_colors[i_ref]: asso_matrix[i_pred, i_ref]
                    for i_ref in range(asso_matrix.shape[1])}
                   for i_pred in range(asso_matrix.shape[0])]
    return asso_colors


def compute_group_overlap_score(ref_labels, pred_labels,
                                threshold_overlap_pred=0.5,
                                threshold_overlap_ref=0.5):
    """How well do the pred_labels explain the ref_labels?

    A predicted cluster explains a reference cluster if it is contained within the reference
    cluster with at least 50% (threshold_overlap_pred) of its points and these correspond
    to at least 50% (threshold_overlap_ref) of the reference cluster.
    """
    ref_unique, ref_counts = np.unique(ref_labels, return_counts=True)
    ref_dict = dict(zip(ref_unique, ref_counts))
    pred_unique, pred_counts = np.unique(pred_labels, return_counts=True)
    pred_dict = dict(zip(pred_unique, pred_counts))
    summary = []
    for true in ref_unique:
        sub_pred_unique, sub_pred_counts = np.unique(pred_labels[true == ref_labels], return_counts=True)
        relative_overlaps_pred = [sub_pred_counts[i] / pred_dict[n] for i, n in enumerate(sub_pred_unique)]
        relative_overlaps_ref = [sub_pred_counts[i] / ref_dict[true] for i, n in enumerate(sub_pred_unique)]
        pred_best_index = np.argmax(relative_overlaps_pred)
        summary.append(1 if (relative_overlaps_pred[pred_best_index] >= threshold_overlap_pred and
                             relative_overlaps_ref[pred_best_index] >= threshold_overlap_ref)
                       else 0)
        # print(true, sub_pred_unique[pred_best_index], relative_overlaps_pred[pred_best_index],
        #       relative_overlaps_ref[pred_best_index], summary[-1])
    return sum(summary)/len(summary)


def identify_groups(ref_labels, pred_labels, return_overlaps=False):
    """Which predicted label explains which reference label?

    A predicted label explains the reference label which maximizes the minimum
    of ``relative_overlaps_pred`` and ``relative_overlaps_ref``.

    Compare this with ``compute_association_matrix_of_groups``.

    Returns
    -------
    A dictionary of length ``len(np.unique(ref_labels))`` that stores for each
    reference label the predicted label that best explains it.

    If ``return_overlaps`` is ``True``, this will in addition return the overlap
    of the reference group with the predicted group; normalized with respect to
    the reference group size and the predicted group size, respectively.
    """
    ref_unique, ref_counts = np.unique(ref_labels, return_counts=True)
    ref_dict = dict(zip(ref_unique, ref_counts))
    pred_unique, pred_counts = np.unique(pred_labels, return_counts=True)
    pred_dict = dict(zip(pred_unique, pred_counts))
    associated_predictions = {}
    associated_overlaps = {}
    for ref_label in ref_unique:
        sub_pred_unique, sub_pred_counts = np.unique(pred_labels[ref_label == ref_labels], return_counts=True)
        relative_overlaps_pred = [sub_pred_counts[i] / pred_dict[n] for i, n in enumerate(sub_pred_unique)]
        relative_overlaps_ref = [sub_pred_counts[i] / ref_dict[ref_label] for i, n in enumerate(sub_pred_unique)]
        relative_overlaps = np.c_[relative_overlaps_pred, relative_overlaps_ref]
        relative_overlaps_min = np.min(relative_overlaps, axis=1)
        pred_best_index = np.argsort(relative_overlaps_min)[::-1]
        associated_predictions[ref_label] = sub_pred_unique[pred_best_index]
        associated_overlaps[ref_label] = relative_overlaps[pred_best_index]
    if return_overlaps: return associated_predictions, associated_overlaps
    else: return associated_predictions


def remove_repetitions_from_list(l):
    return [l[0]] + [e for (i, e) in enumerate(l[1:]) if l[i] != e]


def plot_category_association(adata, prediction, reference, asso_matrix):
    pl.figure(figsize=(5, 5))
    pl.imshow(np.array(asso_matrix)[:], shape=(12, 4))
    pl.xticks(range(len(adata.uns[reference + '_order'])), adata.uns[reference + '_order'], rotation='vertical')
    pl.yticks(range(len(adata.uns[prediction + '_order'])), adata.uns[prediction + '_order'])
    pl.colorbar()


def unique_categories(categories):
    """Pass array-like categories, return sorted cleaned unique categories."""
    categories = np.unique(categories)
    categories = np.setdiff1d(categories, np.array(settings.categories_to_ignore))
    categories = np.array(natsorted(categories, key=lambda v: v.upper()))
    return categories


def fill_in_datakeys(example_parameters, dexdata):
    """Update the 'examples dictionary' _examples.example_parameters.

    If a datakey (key in 'datafile dictionary') is not present in the 'examples
    dictionary' it is used to initialize an entry with that key.

    If not specified otherwise, any 'exkey' (key in 'examples dictionary') is
    used as 'datakey'.
    """
    # default initialization of 'datakey' key with entries from data dictionary
    for exkey in example_parameters:
        if 'datakey' not in example_parameters[exkey]:
            if exkey in dexdata:
                example_parameters[exkey]['datakey'] = exkey
            else:
                example_parameters[exkey]['datakey'] = 'unspecified in dexdata'
    return example_parameters


# backwards compat... remove this in the future
def sanitize_anndata(adata):
    adata._sanitize()


def moving_average(a, n):
    """Moving average over one-dimensional array.

    Parameters
    ----------
    a : np.ndarray
        One-dimensional array.
    n : int
        Number of entries to average over. n=2 means averaging over the currrent
        the previous entry.

    Returns
    -------
    An array view storing the moving average.
    """
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n


# --------------------------------------------------------------------------------
# Deal with tool parameters
# --------------------------------------------------------------------------------


def update_params(old_params, new_params, check=False):
    """Update old_params with new_params.

    If check==False, this merely adds and overwrites the content of old_params.

    If check==True, this only allows updating of parameters that are already
    present in old_params.

    Parameters
    ----------
    old_params : dict
    new_params : dict
    check : bool, optional (default: False)

    Returns
    -------
    updated_params : dict
    """
    updated_params = dict(old_params)
    if new_params:  # allow for new_params to be None
        for key, val in new_params.items():
            if key not in old_params and check:
                raise ValueError('\'' + key
                                 + '\' is not a valid parameter key, '
                                 + 'consider one of \n'
                                 + str(list(old_params.keys())))
            if val is not None:
                updated_params[key] = val
    return updated_params


# --------------------------------------------------------------------------------
# Command-line argument reading and processing
# --------------------------------------------------------------------------------


def read_args_tool(toolkey, example_parameters, tool_add_args=None):
    """Read args for single tool.
    """
    import scanpy as sc
    p = default_tool_argparser(help(toolkey), example_parameters)
    if tool_add_args is None:
        p = add_args(p)
    else:
        p = tool_add_args(p)
    args = vars(p.parse_args())
    args = settings.process_args(args)
    return args


def default_tool_argparser(description, example_parameters):
    """Create default parser for single tools.
    """
    import argparse
    epilog = '\n'
    for k, v in sorted(example_parameters.items()):
        epilog += '  ' + k + '\n'
    p = argparse.ArgumentParser(
        description=description,
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=('available values for examples (exkey):'+epilog))
    return p


# --------------------------------------------------------------------------------
# Others
# --------------------------------------------------------------------------------


def select_groups(adata, groups_order_subset='all', key='groups'):
    """Get subset of groups in adata.obs[key].
    """
    groups_order = adata.obs[key].cat.categories
    if key + '_masks' in adata.uns:
        groups_masks = adata.uns[key + '_masks']
    else:
        groups_masks = np.zeros((len(adata.obs[key].cat.categories),
                                 adata.obs[key].values.size), dtype=bool)
        for iname, name in enumerate(adata.obs[key].cat.categories):
            # if the name is not found, fallback to index retrieval
            if adata.obs[key].cat.categories[iname] in adata.obs[key].values:
                mask = adata.obs[key].cat.categories[iname] == adata.obs[key].values
            else:
                mask = str(iname) == adata.obs[key].values
            groups_masks[iname] = mask
    groups_ids = list(range(len(groups_order)))
    if groups_order_subset != 'all':
        groups_ids = []
        for name in groups_order_subset:
            groups_ids.append(
                np.where(adata.obs[key].cat.categories.values == name)[0][0])
        if len(groups_ids) == 0:
            # fallback to index retrieval
            groups_ids = np.where(
                np.in1d(np.arange(len(adata.obs[key].cat.categories)).astype(str),
                                          np.array(groups_order_subset)))[0]
        if len(groups_ids) == 0:
            logg.m(np.array(groups_order_subset),
                   'invalid! specify valid groups_order (or indices) one of',
                   adata.obs[key].cat.categories)
            from sys import exit
            exit(0)
        groups_masks = groups_masks[groups_ids]
        groups_order_subset = adata.obs[key].cat.categories[groups_ids].values
    else:
        groups_order_subset = groups_order.values
    return groups_order_subset, groups_masks


def pretty_dict_string(d, indent=0):
    """Pretty output of nested dictionaries.
    """
    s = ''
    for key, value in sorted(d.items()):
        s += '    ' * indent + str(key)
        if isinstance(value, dict):
             s += '\n' + pretty_dict_string(value, indent+1)
        else:
             s += '=' + str(value) + '\n'
    return s


def markdown_dict_string(d):
    """Markdown output that can be pasted in the examples/README.md.
    """
    # sort experimental data from simulated data
    from collections import OrderedDict
    types = OrderedDict()
    for key, value in sorted(d.items()):
        if 'type' in value:
            if value['type'] not in types:
                types[value['type']] = []
            types[value['type']].append(key)
        else:
            print(key, 'does not define data type!')
    # format output
    s = ''
    for type in ['scRNAseq', 'scqPCR', 'bulk', 'simulated']:
        s += '\nExamples using ' + type + ' data.\n'
        for key in types[type]:
            value = d[key]
            s += '* [' + key + '](#' + key + ')'
            if 'ref' in value:
                if 'doi' in value:
                    link = 'http://dx.doi.org/' + value['doi']
                elif 'url' in value:
                    link = value['url']
                s += (' - [' +  value['ref'].replace('et al.','*et al.*')
                             + '](' + link +  ')')
            if 'title' in value:
                s += '   \n*' + value['title'] + '*'
            s += '\n'
    return s


def merge_dicts(*ds):
    """Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.

    Notes
    -----
    http://stackoverflow.com/questions/38987/how-to-merge-two-python-dictionaries-in-a-single-expression
    """
    result = ds[0]
    for d in ds[1:]:
        result.update(d)
    return result


def masks(list_of_index_lists, n):
    """Make an array in which rows store 1d mask arrays from list of index lists.

    Parameters
    ----------
    n : int
        Maximal index / number of samples.
    """
    # make a list of mask arrays, it's easier to store
    # as there is a hdf5 equivalent
    for il,l in enumerate(list_of_index_lists):
        mask = np.zeros(n,dtype=bool)
        mask[l] = True
        list_of_index_lists[il] = mask
    # convert to arrays
    masks = np.array(list_of_index_lists)
    return masks


def warn_with_traceback(message, category, filename, lineno, file=None, line=None):
    """Get full tracebacks when warning is raised by setting

    warnings.showwarning = warn_with_traceback

    See also
    --------
    http://stackoverflow.com/questions/22373927/get-traceback-of-warnings
    """
    import warnings
    import traceback
    traceback.print_stack()
    log = file if hasattr(file, 'write') else sys.stderr
    settings.write(warnings.formatwarning(message, category, filename, lineno, line))


def subsample(X, subsample=1, seed=0):
    """Subsample a fraction of 1/subsample samples from the rows of X.

    Parameters
    ----------
    X : np.ndarray
        Data array.
    subsample : int
        1/subsample is the fraction of data sampled, n = X.shape[0]/subsample.
    seed : int
        Seed for sampling.

    Returns
    -------
    Xsampled : np.ndarray
        Subsampled X.
    rows : np.ndarray
        Indices of rows that are stored in Xsampled.
    """
    if subsample == 1 and seed == 0:
        return X, np.arange(X.shape[0], dtype=int)
    if seed == 0:
        # this sequence is defined simply by skipping rows
        # is faster than sampling
        rows = np.arange(0, X.shape[0], subsample, dtype=int)
        n = rows.size
        Xsampled = np.array(X[rows])
    if seed > 0:
        n = int(X.shape[0]/subsample)
        np.random.seed(seed)
        Xsampled, rows = subsample_n(X, n=n)
    logg.m('... subsampled to', n, 'of', X.shape[0], 'data points')
    return Xsampled, rows


def subsample_n(X, n=0, seed=0):
    """Subsample n samples from rows of array.

    Parameters
    ----------
    X : np.ndarray
        Data array.
    seed : int
        Seed for sampling.

    Returns
    -------
    Xsampled : np.ndarray
        Subsampled X.
    rows : np.ndarray
        Indices of rows that are stored in Xsampled.
    """
    if n < 0:
        raise ValueError('n must be greater 0')
    np.random.seed(seed)
    n = X.shape[0] if (n == 0 or n > X.shape[0]) else n
    rows = np.random.choice(X.shape[0], size=n, replace=False)
    Xsampled = X[rows]
    return Xsampled, rows


def check_presence_download(filename, backup_url):
    """Check if file is present otherwise download."""
    import os
    if not os.path.exists(filename):
        from .readwrite import download_progress
        dr = os.path.dirname(filename)
        try:
            os.makedirs(dr)
        except FileExistsError:
            pass  # ignore if dir already exists
        from urllib.request import urlretrieve
        urlretrieve(backup_url, filename, reporthook=download_progress)


def hierarch_cluster(M):
    """Cluster matrix using hierarchical clustering.

    Parameters
    ----------
    M : np.ndarray
        Matrix, for example, distance matrix.

    Returns
    -------
    Mclus : np.ndarray
        Clustered matrix.
    indices : np.ndarray
        Indices used to cluster the matrix.
    """
    import scipy as sp
    import scipy.cluster
    link = sp.cluster.hierarchy.linkage(M)
    indices = sp.cluster.hierarchy.leaves_list(link)
    Mclus = np.array(M[:, indices])
    Mclus = Mclus[indices, :]
    if False:
        pl.matshow(Mclus)
        pl.colorbar()
    return Mclus, indices



import anndata as ad

def readandimputematrix(file_name, min_coverage=1):
    """
    Temporary function to load and impute methyaltion count matrix into an AnnData object
    
    Parameters
    ----------
    file_name : file name to read and load
    
    min_coverage : minimum number of cells covered for which we keep and impute a variable
    
    Returns
    -------
    adata : :class:`~anndata.AnnData`
        Annotated data matrix.
    
    """
    with open(file_name) as f:
        file = f.readlines()

    # separate annotation from data    
    head_var = file[0]
    head_var = head_var.split('\t')
    # Then, extract the sample names
    sample_names = []
    data_raw = []
    for l in file[1:]:
        l = l.split('\t')
        sample_names.append(l[0])
        data_raw.append(l[1:])

    # clear memory of useless variables 
    del file
    
    ##########################################
    # now, removing empty columns
    empties = []
    partial = []
    full =  []
    for index in range(1, len(data_raw[0])):
        column = [element[index] for element in data_raw]
        nan_count =0
       	for i in column:
       		if i == 'nan':
       			nan_count +=1
        if len(list(set(column))) == 1:
            empties.append(index)
        #elif len(list(set(column))) <= min_coverage:
        #    partial.append(index)
        elif len(column)-nan_count <= min_coverage:
            partial.append(index)
        else:
            full.append(index)
         
    ##########################################
    intermed_matrix = []
    name_windows_covered = []
    # let's remove the compltetly uninformative columns
    for index in range(1, len(head_var[1:])):
        if index in full:
            intermed_matrix.append([element[index] for element in data_raw])
            name_windows_covered.append(head_var[index])

    ########################################
    # imputing values.
    imputed_matrix = []
    for row in intermed_matrix:
        imputed_row = []
        if "nan" in row:
            mean = np.mean([float(e)  for e in row if e != "nan"])
            for element in row:
                if element == "nan":
                    imputed_row.append(str(mean))
                else: 
                    imputed_row.append(element)
            imputed_matrix.append(imputed_row)
        else:
            imputed_matrix.append(row)
        
    ########################################
     # return adata object.
    
    #imputed_matrix = np.matrix(imputed_matrix).transpose()
    #return(imputed_matrix, sample_names, name_windows_covered)
    adata = ad.AnnData(np.matrix(imputed_matrix),
                       obs=pd.DataFrame(index=name_windows_covered),
                       var=pd.DataFrame(index=sample_names))
    adata.transpose()
    return(adata)



# chromosomes for 2 principal species. If you work with another genome
# the chromosomes will have to be specified
HUMAN = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '2',
         '20', '21', '22', '3', '4', '5', '6', '7', '8', '9', 'X', 'Y']
MOUSE = ['1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '2',
         '3', '4', '5', '6', '7', '8', '9', 'X', 'Y']
# it assume that you take a bed file as impute
# it save the features position and name like [start, end, name]
# in a dictionary that has chromosomes as keys. 
def load_features(file_features, chromosomes=HUMAN, path="", sort=False):
    """
    Description function. 
    
    The function load features is here to transform a bed file into a usable 
    set of units to measure methylation levels. 
    It has to be a bed-like file. You also need to specify the chromosomes you 
    use as a list of characteres like ['1', '7', '8', '9', 'X', 'Y', 'M'].
    The chromosome list you give as input can be not ordered.
    If you don't specify the chromosomes, the default is the human genome
    (including X, Y and mitochondrial DNA).
    THE BED FILE need to be sorted !! (Maybe I should add an option to sort the
    file).
    
    The output is a dictionary where the keys are chromosomes and the value is
    a list containing [start, end, name] for every feature extracted.
    
    Parameters
    ----------
    file_features:
        the names of the bed file you want to load.
    chromosomes:
        chromosomes corresponding to the bed file. 
        If not specified, it's human by default
    path:
        if you want to specify the path where your bed file is. 
    sort: 
        if True, the bed file is sorted based on starting coordinates.
    """
    features_chrom = {}
    for c in chromosomes:
        features_chrom[c] = []
    with open(path + file_features) as features:
        for line in features:
            line = line.split("\t")
            if line[0][3:] in chromosomes:
                features_chrom[line[0][3:]].append([int(line[1]), int(line[2]), line[3]])
                
#    for c in chromosomes:
#        if features_chrom[c] ==  []:
#            del features_chrom[c]
#            
    if sort == True:
        for c in chromosomes:
            sorted(features_chrom[c], key=lambda x: x[0])
            
    return(features_chrom)


def make_windows(size, chromosomes=HUMAN, max_length=1000000000):
    """
    Description function. 
    
    Generate windows/bins of the given size for the appropriate genome (default
    choice is human). 
    
    Parameters
    ----------
    size:
        size of the window you want
    chromosomes:
        Chromosomes of the species you are analysing
    max_length:
        maximum length given for every chromosome
    """
    features_chrom = {}
    start = range(0, max_length - size, size)
    end = range(size, max_length, size)
    for c in chromosomes:
        features_chrom[c] = [[start[i], end[i], ''.join(["chr_", c, "_", str(i)])] for i in range(len(end))]
    return(features_chrom)



import scanpy.api as sc
import anndata as ad

def lazy(adata, pp_pca=True, copy=False):
    '''
    Automatically computes PCA coordinates, loadings and variance decomposition, a neighborhood graph of observations,
    t-distributed stochastic neighborhood embedding (tSNE) Uniform Manifold Approximation and Projection (UMAP)
    
    Parameters
    ----------
    adata : :class:`~anndata.AnnData`
        Annotated data matrix.
    pp_pca :  `bool` (default: `True`)
        Computes PCA coordinates before the neighborhood graph
    copy : `bool` (default: `False`)
        Return a copy instead of writing to adata.
    
    
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    X_pca : `adata.obsm`
        PCA coordinates of data.
    connectivities : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Weighted adjacency matrix of the neighborhood graph of data
        points. Weights should be interpreted as connectivities.
    distances : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Instead of decaying weights, this stores distances for each pair of
        neighbors.
    X_tsne : `np.ndarray` (`adata.obs`, dtype `float`)
        tSNE coordinates of data.
    X_umap : `adata.obsm`
        UMAP coordinates of data.
    
    
    '''
    if copy:
        adata = adata.copy() 
    else:
        adata

    if pp_pca:
        sc.pp.pca(adata)
    
    sc.pp.neighbors(adata)
    sc.tl.pca(adata)
    sc.tl.tsne(adata)
    sc.tl.umap(adata)
    
    if copy:
        return(adata)
    else:
        None



import anndata as ad

def top_feature_genes(adata, gtf_file, extension=5000):
    
    # extract_top_markers
    #print(adata2.uns['rank_genes_groups'].keys())
    windows = [list(w) for w in adata2.uns['rank_genes_groups']['names'].tolist()]
    windows_all = []
    for w in windows:
        windows_all += w

    # load the gtf file
    with open(gtf_file) as f:
        gtf_raw = f.readlines()
    gtf = []
    for line in gtf_raw:
        if line[0] != '#':
            gtf.append(line[:-2].split('\t'))
    del gtf_raw

    markers = []
    for w in windows_all:
        curr_m = []
        w2 = w.split('_')
        chrom = w2[0][3:]
        w2 = [int(x) for x in w2[1:]]
        for gene in gtf_dic[chrom]:
            start = int(gene[3])-EXTENSION
            end = int(gene[4])+EXTENSION
            if (w2[0] < start < w2[1]) or (w2[0] < end < w2[1]):
                gene_name = gene[-1]
                for n in gene[-1].split(';'):
                    if 'gene_name' in n:
                        gene_name = n
                curr_m.append([w, gene_name])
        if curr_m != []:
            markers += curr_m
    markers = [list(x) for x in set(tuple(x) for x in markers)]
    markers_dict={}
    for n in markers:
        markers_dict[n[1].split(' "')[1][:-1]] = n[0]
    return(markers_dict)



import numpy as np
import anndata as ad

def load_met_noimput(matrix_file, path='', save=False):
    """
    read the raw count matrix and convert it into an AnnData object.
    write down the matrix as .h5ad (AnnData object) if save = True.
    Return AnnData object
    """
    matrix = []
    cell_names = []
    feature_names = []
    with open('../../enhancer_c1_CG_paper_unique_corrected2.txt') as f:
        line = f.readline()[:-2].split('\t')
        if line[0] == 'sample_name':
            feature_names = line[1:-1]
        else:
            matrix.append(line[1:])
            cell_names.append(line[0])
        if matrix == []:
            line = f.readline()[:-2].split('\t')
            matrix.append(line[1:])
            cell_names.append(line[0])
        for line in f:
            line = line[:-2].split('\t')
            matrix.append(line[1:])
            cell_names.append(line[0])

    matrix = np.array(matrix)
    
    if feature_names != []:
        adata = ad.AnnData(matrix, obs=pd.DataFrame(index=cell_names),  var=pd.DataFrame(index=feature_names))
    else:
        adata = ad.AnnData(matrix, obs=pd.DataFrame(index=cell_names))
    
    adata.uns['omic'] = 'methylation'
    
    if save:
        adata.write("".join([".".split(matrix_file)[0],'.h5ad']))
        
    return(adata)

def imputation_met(adata, number_cell_covered=10, imputation_value='mean', save=None):
    """
    Impute missing values in methyaltion level matrices. The imputsation is based on the average
    methylation value of the given variable.
    It also filter out variables that are covered in an unsufficient number of cells in order to 
    reduce the feature space to meaningful variables and discard potential coverage biases. 

    Parameters
    ----------
    adata: AnnData object containing 'nan'

    number_cell_covered: minimum number of cells to be covered in order to retain a variable

    imputation_value: imputation of the missing value can be made either on the mean or the median

    Return
    ------
    Return a new AnnData object

    
    
    """

    # This step need to be sped up and coule be multithread.
    # Only the mean available for now. And only the minimum number of cells covered and not the variety of the 
    # methylation levels
    # also, it odes not return the variable annoations and force to add 2 values
    matrix = adata.X.transpose().tolist()
    old_features = adata.var_names.tolist()
    
    new_matrix = []
    new_features_name = []
    means = []
    medians = []
    feat_nb = 0
    for feature in adata.X:
        nb_cov = 0
        # also the different number of values
        val_for_mean = []
        for f in feature:
            if 0.0 <= f <= 1.0:
                nb_cov += 1
                val_for_mean.append(f)
    
    
        new_feature = []
    
        if nb_cov >= number_cell_covered:
            new_features_name.append(old_features[feat_nb])
            # also median possible
            mean = np.mean(val_for_mean)
            for f in feature:
                if 0.0 <= f <= 1.0:
                    new_feature.append(f)
                else:
                    new_feature.append(mean)
                
            new_matrix.append(new_feature)
        
            means.append(mean)
            medians.append(np.median(val_for_mean))
            
        feat_nb += 1
    
    
    new_matrix = np.array(new_matrix)
    return(new_matrix, new_features_name)
    adata2 = ad.AnnData(new_matrix.transpose(), obs=adata.obs_names, var = pd.DataFrame(index=new_features_name))
    adata2.obs = adata.obs
    adata2.var['median_before_imput'] = medians # before imputation
    adata2.var['mean_before_imput'] = means # before imputation
    
    if save == None:
        return(adata2)
    else:
        if (len(save)>=5) and (save[-5:] == '.h5ad'):
            adata2.write(save)
        else:
            adata2.write(save+'.h5ad')
        return(adata2)

def readandimputematrix(file_name, min_coverage=1):
    """
    Temporary function to load and impute methyaltion count matrix into an AnnData object
    
    Parameters
    ----------
    file_name : file name to read and load
    
    min_coverage : minimum number of cells covered for which we keep and impute a variable
    
    Returns
    -------
    adata : :class:`~anndata.AnnData`
        Annotated data matrix.
    
    """
    with open(file_name) as f:
        file = f.readlines()

    # separate annotation from data    
    head_var = file[0]
    head_var = head_var.split('\t')
    # Then, extract the sample names
    sample_names = []
    data_raw = []
    for l in file[1:]:
        l = l.split('\t')
        sample_names.append(l[0])
        data_raw.append(l[1:])

    # clear memory of useless variables 
    del file
    
    ##########################################
    # now, removing empty columns
    empties = []
    partial = []
    full =  []
    for index in range(1, len(data_raw[0])):
        column = [element[index] for element in data_raw]
        if len(list(set(column))) == 1:
            empties.append(index)
        elif len(list(set(column))) <= min_coverage:
            partial.append(index)
        else:
            full.append(index)
         
    ##########################################
    intermed_matrix = []
    name_windows_covered = []
    # let's remove the compltetly uninformative columns
    for index in range(1, len(head_var[1:])):
        if index in full:
            intermed_matrix.append([element[index] for element in data_raw])
            name_windows_covered.append(head_var[index])

    ########################################
    # imputing values.
    imputed_matrix = []
    for row in intermed_matrix:
        imputed_row = []
        if "nan" in row:
            mean = np.mean([float(e)  for e in row if e != "nan"])
            for element in row:
                if element == "nan":
                    imputed_row.append(str(mean))
                else: 
                    imputed_row.append(element)
            imputed_matrix.append(imputed_row)
        else:
            imputed_matrix.append(row)

    imputed_matrix = np.matrix(imputed_matrix).transpose()
    return(imputed_matrix, sample_names, name_windows_covered)
